# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ContractInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'contract_type': 'str',
        'starting_at': 'datetime',
        'ending_before': 'datetime',
        'is_archived': 'bool'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'contract_type': 'contract_type',
        'starting_at': 'starting_at',
        'ending_before': 'ending_before',
        'is_archived': 'is_archived'
    }

    def __init__(self, id=None, name=None, contract_type=None, starting_at=None, ending_before=None, is_archived=False, local_vars_configuration=None):  # noqa: E501
        """ContractInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._contract_type = None
        self._starting_at = None
        self._ending_before = None
        self._is_archived = None
        self.discriminator = None

        self.id = id
        if name is not None:
            self.name = name
        if contract_type is not None:
            self.contract_type = contract_type
        if starting_at is not None:
            self.starting_at = starting_at
        if ending_before is not None:
            self.ending_before = ending_before
        if is_archived is not None:
            self.is_archived = is_archived

    @property
    def id(self):
        """Gets the id of this ContractInfo.  # noqa: E501

        Contract ID  # noqa: E501

        :return: The id of this ContractInfo.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ContractInfo.

        Contract ID  # noqa: E501

        :param id: The id of this ContractInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this ContractInfo.  # noqa: E501

        Contract name  # noqa: E501

        :return: The name of this ContractInfo.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ContractInfo.

        Contract name  # noqa: E501

        :param name: The name of this ContractInfo.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def contract_type(self):
        """Gets the contract_type of this ContractInfo.  # noqa: E501

        Contract type  # noqa: E501

        :return: The contract_type of this ContractInfo.  # noqa: E501
        :rtype: str
        """
        return self._contract_type

    @contract_type.setter
    def contract_type(self, contract_type):
        """Sets the contract_type of this ContractInfo.

        Contract type  # noqa: E501

        :param contract_type: The contract_type of this ContractInfo.  # noqa: E501
        :type: str
        """

        self._contract_type = contract_type

    @property
    def starting_at(self):
        """Gets the starting_at of this ContractInfo.  # noqa: E501

        Contract start date  # noqa: E501

        :return: The starting_at of this ContractInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._starting_at

    @starting_at.setter
    def starting_at(self, starting_at):
        """Sets the starting_at of this ContractInfo.

        Contract start date  # noqa: E501

        :param starting_at: The starting_at of this ContractInfo.  # noqa: E501
        :type: datetime
        """

        self._starting_at = starting_at

    @property
    def ending_before(self):
        """Gets the ending_before of this ContractInfo.  # noqa: E501

        Contract end date  # noqa: E501

        :return: The ending_before of this ContractInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._ending_before

    @ending_before.setter
    def ending_before(self, ending_before):
        """Sets the ending_before of this ContractInfo.

        Contract end date  # noqa: E501

        :param ending_before: The ending_before of this ContractInfo.  # noqa: E501
        :type: datetime
        """

        self._ending_before = ending_before

    @property
    def is_archived(self):
        """Gets the is_archived of this ContractInfo.  # noqa: E501

        Whether contract is archived  # noqa: E501

        :return: The is_archived of this ContractInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_archived

    @is_archived.setter
    def is_archived(self, is_archived):
        """Sets the is_archived of this ContractInfo.

        Whether contract is archived  # noqa: E501

        :param is_archived: The is_archived of this ContractInfo.  # noqa: E501
        :type: bool
        """

        self._is_archived = is_archived

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ContractInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ContractInfo):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass, field
from datetime import datetime

from anyscale._private.models import ModelBase


@dataclass(frozen=True)
class UserGroup(ModelBase):
    """A user group in the organization."""

    __doc_py_example__ = """\
import anyscale

user_group = anyscale.user_group.get(id="ug_abc123")
print(f"{user_group.id}: {user_group.name}")
"""

    id: str = field(metadata={"docstring": "The unique identifier of the user group."})

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("id must be a string.")

    name: str = field(metadata={"docstring": "The name of the user group."})

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("name must be a string.")

    org_id: str = field(metadata={"docstring": "The organization ID."})

    def _validate_org_id(self, org_id: str):
        if not isinstance(org_id, str):
            raise TypeError("org_id must be a string.")

    created_at: datetime = field(
        metadata={"docstring": "When the user group was created."}
    )

    def _validate_created_at(self, created_at: datetime):
        if not isinstance(created_at, datetime):
            raise TypeError("created_at must be a datetime.")

    updated_at: datetime = field(
        metadata={"docstring": "When the user group was last updated."}
    )

    def _validate_updated_at(self, updated_at: datetime):
        if not isinstance(updated_at, datetime):
            raise TypeError("updated_at must be a datetime.")

    @classmethod
    def from_api_model(cls, api_model) -> "UserGroup":
        """Create a UserGroup from an API model."""
        return cls(
            id=api_model.id,
            name=api_model.name,
            org_id=api_model.org_id,
            created_at=api_model.created_at,
            updated_at=api_model.updated_at,
        )



import sys
from typing import Optional

from packaging import version
import requests

import anyscale
from anyscale.util import is_anyscale_cluster, is_anyscale_workspace


ANYSCALE_PYPI_PACKAGE = "anyscale"
VERSION_UPGRADE_THRESHOLD = 5


def _get_latest_pypi_version(package_name: str) -> Optional[str]:
    """
    This method gets the latest version of a package from PyPI.
    If the request fails, it doesn't raise the execption since it shouldn't
    interrupt the CLI flow.
    """
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json", timeout=2)
        data = response.json()
        return data["info"]["version"]
    except Exception:  # noqa: BLE001
        return None


def _is_upgrade_needed(
    local_anyscale_version: str,
    latest_version: Optional[str],
    version_threshold: int = VERSION_UPGRADE_THRESHOLD,
) -> bool:
    """
    Check if the local Anyscale version needs an upgrade.

    Returns True if:
    - The major or minor version is behind the latest version, OR
    - The patch version is more than `version_threshold` versions behind.

    This reduces noise from frequent patch releases while still warning users
    about significant version differences.
    """
    if local_anyscale_version == "0.0.0-dev":
        # This is using anyscale CLI from source. No need to check for updates.
        return False
    if latest_version is None:
        return False

    local_ver = version.parse(local_anyscale_version)
    latest_ver = version.parse(latest_version)

    if local_ver >= latest_ver:
        return False

    # Always warn if major or minor version is behind
    if local_ver.major < latest_ver.major or local_ver.minor < latest_ver.minor:
        return True

    # Only warn if more than version_threshold patch versions behind
    return latest_ver.micro - local_ver.micro > version_threshold


def log_warning_if_version_needs_upgrade() -> None:
    """
    Logs a warning if the local Anyscale version is not the latest.

    Note that the check is skipped for clusters/workspaces.
    """
    if is_anyscale_workspace() or is_anyscale_cluster():
        # If the command is run from clusters/workspaces, we don't do version checking
        # because the Anyscale CLI is bundled as a part of cluster envs.
        return

    local_anyscale_version = anyscale.__version__
    latest_version = _get_latest_pypi_version(ANYSCALE_PYPI_PACKAGE)

    if _is_upgrade_needed(local_anyscale_version, latest_version):
        print(
            "[WARNING] A newer version of the Anyscale CLI is available. "
            f"Your current version is {local_anyscale_version}. The latest version is {latest_version}. "
            "To avoid issues accessing Anyscale`s API, upgrade to the latest version by running `pip install --upgrade anyscale`.",
            file=sys.stderr,
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreditsV2(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'in_use_credits': 'list[CreditRecordCreditV2]',
        'pending_credits': 'list[CreditRecordCreditV2]',
        'expired_credits': 'list[CreditRecordCreditV2]',
        'in_use_commits': 'list[CreditRecordCommitV2]',
        'pending_commits': 'list[CreditRecordCommitV2]',
        'expired_commits': 'list[CreditRecordCommitV2]',
        'current_balance_usd': 'float',
        'amount_spent_usd': 'float',
        'total_granted_usd': 'float'
    }

    attribute_map = {
        'in_use_credits': 'in_use_credits',
        'pending_credits': 'pending_credits',
        'expired_credits': 'expired_credits',
        'in_use_commits': 'in_use_commits',
        'pending_commits': 'pending_commits',
        'expired_commits': 'expired_commits',
        'current_balance_usd': 'current_balance_usd',
        'amount_spent_usd': 'amount_spent_usd',
        'total_granted_usd': 'total_granted_usd'
    }

    def __init__(self, in_use_credits=None, pending_credits=None, expired_credits=None, in_use_commits=None, pending_commits=None, expired_commits=None, current_balance_usd=None, amount_spent_usd=None, total_granted_usd=None, local_vars_configuration=None):  # noqa: E501
        """CreditsV2 - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._in_use_credits = None
        self._pending_credits = None
        self._expired_credits = None
        self._in_use_commits = None
        self._pending_commits = None
        self._expired_commits = None
        self._current_balance_usd = None
        self._amount_spent_usd = None
        self._total_granted_usd = None
        self.discriminator = None

        self.in_use_credits = in_use_credits
        self.pending_credits = pending_credits
        self.expired_credits = expired_credits
        self.in_use_commits = in_use_commits
        self.pending_commits = pending_commits
        self.expired_commits = expired_commits
        self.current_balance_usd = current_balance_usd
        self.amount_spent_usd = amount_spent_usd
        self.total_granted_usd = total_granted_usd

    @property
    def in_use_credits(self):
        """Gets the in_use_credits of this CreditsV2.  # noqa: E501

        List of in-use credits.  # noqa: E501

        :return: The in_use_credits of this CreditsV2.  # noqa: E501
        :rtype: list[CreditRecordCreditV2]
        """
        return self._in_use_credits

    @in_use_credits.setter
    def in_use_credits(self, in_use_credits):
        """Sets the in_use_credits of this CreditsV2.

        List of in-use credits.  # noqa: E501

        :param in_use_credits: The in_use_credits of this CreditsV2.  # noqa: E501
        :type: list[CreditRecordCreditV2]
        """
        if self.local_vars_configuration.client_side_validation and in_use_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `in_use_credits`, must not be `None`")  # noqa: E501

        self._in_use_credits = in_use_credits

    @property
    def pending_credits(self):
        """Gets the pending_credits of this CreditsV2.  # noqa: E501

        List of pending credits.  # noqa: E501

        :return: The pending_credits of this CreditsV2.  # noqa: E501
        :rtype: list[CreditRecordCreditV2]
        """
        return self._pending_credits

    @pending_credits.setter
    def pending_credits(self, pending_credits):
        """Sets the pending_credits of this CreditsV2.

        List of pending credits.  # noqa: E501

        :param pending_credits: The pending_credits of this CreditsV2.  # noqa: E501
        :type: list[CreditRecordCreditV2]
        """
        if self.local_vars_configuration.client_side_validation and pending_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `pending_credits`, must not be `None`")  # noqa: E501

        self._pending_credits = pending_credits

    @property
    def expired_credits(self):
        """Gets the expired_credits of this CreditsV2.  # noqa: E501

        List of expired credits.  # noqa: E501

        :return: The expired_credits of this CreditsV2.  # noqa: E501
        :rtype: list[CreditRecordCreditV2]
        """
        return self._expired_credits

    @expired_credits.setter
    def expired_credits(self, expired_credits):
        """Sets the expired_credits of this CreditsV2.

        List of expired credits.  # noqa: E501

        :param expired_credits: The expired_credits of this CreditsV2.  # noqa: E501
        :type: list[CreditRecordCreditV2]
        """
        if self.local_vars_configuration.client_side_validation and expired_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `expired_credits`, must not be `None`")  # noqa: E501

        self._expired_credits = expired_credits

    @property
    def in_use_commits(self):
        """Gets the in_use_commits of this CreditsV2.  # noqa: E501

        List of in-use commits.  # noqa: E501

        :return: The in_use_commits of this CreditsV2.  # noqa: E501
        :rtype: list[CreditRecordCommitV2]
        """
        return self._in_use_commits

    @in_use_commits.setter
    def in_use_commits(self, in_use_commits):
        """Sets the in_use_commits of this CreditsV2.

        List of in-use commits.  # noqa: E501

        :param in_use_commits: The in_use_commits of this CreditsV2.  # noqa: E501
        :type: list[CreditRecordCommitV2]
        """
        if self.local_vars_configuration.client_side_validation and in_use_commits is None:  # noqa: E501
            raise ValueError("Invalid value for `in_use_commits`, must not be `None`")  # noqa: E501

        self._in_use_commits = in_use_commits

    @property
    def pending_commits(self):
        """Gets the pending_commits of this CreditsV2.  # noqa: E501

        List of pending commits.  # noqa: E501

        :return: The pending_commits of this CreditsV2.  # noqa: E501
        :rtype: list[CreditRecordCommitV2]
        """
        return self._pending_commits

    @pending_commits.setter
    def pending_commits(self, pending_commits):
        """Sets the pending_commits of this CreditsV2.

        List of pending commits.  # noqa: E501

        :param pending_commits: The pending_commits of this CreditsV2.  # noqa: E501
        :type: list[CreditRecordCommitV2]
        """
        if self.local_vars_configuration.client_side_validation and pending_commits is None:  # noqa: E501
            raise ValueError("Invalid value for `pending_commits`, must not be `None`")  # noqa: E501

        self._pending_commits = pending_commits

    @property
    def expired_commits(self):
        """Gets the expired_commits of this CreditsV2.  # noqa: E501

        List of expired commits.  # noqa: E501

        :return: The expired_commits of this CreditsV2.  # noqa: E501
        :rtype: list[CreditRecordCommitV2]
        """
        return self._expired_commits

    @expired_commits.setter
    def expired_commits(self, expired_commits):
        """Sets the expired_commits of this CreditsV2.

        List of expired commits.  # noqa: E501

        :param expired_commits: The expired_commits of this CreditsV2.  # noqa: E501
        :type: list[CreditRecordCommitV2]
        """
        if self.local_vars_configuration.client_side_validation and expired_commits is None:  # noqa: E501
            raise ValueError("Invalid value for `expired_commits`, must not be `None`")  # noqa: E501

        self._expired_commits = expired_commits

    @property
    def current_balance_usd(self):
        """Gets the current_balance_usd of this CreditsV2.  # noqa: E501

        Summary of current balance of all credits and commits.  # noqa: E501

        :return: The current_balance_usd of this CreditsV2.  # noqa: E501
        :rtype: float
        """
        return self._current_balance_usd

    @current_balance_usd.setter
    def current_balance_usd(self, current_balance_usd):
        """Sets the current_balance_usd of this CreditsV2.

        Summary of current balance of all credits and commits.  # noqa: E501

        :param current_balance_usd: The current_balance_usd of this CreditsV2.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and current_balance_usd is None:  # noqa: E501
            raise ValueError("Invalid value for `current_balance_usd`, must not be `None`")  # noqa: E501

        self._current_balance_usd = current_balance_usd

    @property
    def amount_spent_usd(self):
        """Gets the amount_spent_usd of this CreditsV2.  # noqa: E501

        Summary of total amount spent across all credits and commits.  # noqa: E501

        :return: The amount_spent_usd of this CreditsV2.  # noqa: E501
        :rtype: float
        """
        return self._amount_spent_usd

    @amount_spent_usd.setter
    def amount_spent_usd(self, amount_spent_usd):
        """Sets the amount_spent_usd of this CreditsV2.

        Summary of total amount spent across all credits and commits.  # noqa: E501

        :param amount_spent_usd: The amount_spent_usd of this CreditsV2.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and amount_spent_usd is None:  # noqa: E501
            raise ValueError("Invalid value for `amount_spent_usd`, must not be `None`")  # noqa: E501

        self._amount_spent_usd = amount_spent_usd

    @property
    def total_granted_usd(self):
        """Gets the total_granted_usd of this CreditsV2.  # noqa: E501

        Summary of total granted across all credits and commits.  # noqa: E501

        :return: The total_granted_usd of this CreditsV2.  # noqa: E501
        :rtype: float
        """
        return self._total_granted_usd

    @total_granted_usd.setter
    def total_granted_usd(self, total_granted_usd):
        """Sets the total_granted_usd of this CreditsV2.

        Summary of total granted across all credits and commits.  # noqa: E501

        :param total_granted_usd: The total_granted_usd of this CreditsV2.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and total_granted_usd is None:  # noqa: E501
            raise ValueError("Invalid value for `total_granted_usd`, must not be `None`")  # noqa: E501

        self._total_granted_usd = total_granted_usd

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreditsV2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreditsV2):
            return True

        return self.to_dict() != other.to_dict()



from typing import List, Optional

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk import sdk_command
from anyscale.project._private.project_sdk import PrivateProjectSDK
from anyscale.project.models import (
    CreateProjectCollaborator,
    Project,
    ProjectSortField,
    ProjectSortOrder,
)


_PROJECT_SDK_SINGLETON_KEY = "project_sdk"

_ADD_COLLABORATORS_EXAMPLE = """
import anyscale
from anyscale.project.models import CreateProjectCollaborator, ProjectPermissionLevel

anyscale.project.add_collaborators(
    cloud="cloud_name",
    project="project_name",
    collaborators=[
        CreateProjectCollaborator(
            email="test1@anyscale.com",
            permission_level=ProjectPermissionLevel.OWNER,
        ),
        CreateProjectCollaborator(
            email="test2@anyscale.com",
            permission_level=ProjectPermissionLevel.WRITE,
        ),
        CreateProjectCollaborator(
            email="test3@anyscale.com",
            permission_level=ProjectPermissionLevel.READONLY,
        ),
    ],
)
"""

_ADD_COLLABORATORS_DOCSTRINGS = {
    "cloud": "The cloud that the project belongs to.",
    "project": "The project to add users to.",
    "collaborators": "The list of collaborators to add to the project.",
}


@sdk_command(
    _PROJECT_SDK_SINGLETON_KEY,
    PrivateProjectSDK,
    doc_py_example=_ADD_COLLABORATORS_EXAMPLE,
    arg_docstrings=_ADD_COLLABORATORS_DOCSTRINGS,
)
def add_collaborators(
    cloud: str,
    project: str,
    collaborators: List[CreateProjectCollaborator],
    *,
    _private_sdk: Optional[PrivateProjectSDK] = None
) -> str:
    """Batch add collaborators to a project.
    """
    return _private_sdk.add_collaborators(cloud, project, collaborators)  # type: ignore


_GET_PROJECT_EXAMPLE = """
import anyscale
from anyscale.project.models import Project

project: Project = anyscale.project.get(project_id="my-project-id")
"""

_GET_PROJECT_DOCSTRINGS = {
    "project_id": "The ID of the project to get details of.",
}


@sdk_command(
    _PROJECT_SDK_SINGLETON_KEY,
    PrivateProjectSDK,
    doc_py_example=_GET_PROJECT_EXAMPLE,
    arg_docstrings=_GET_PROJECT_DOCSTRINGS,
)
def get(
    project_id: str, *, _private_sdk: Optional[PrivateProjectSDK] = None
) -> Project:
    """Get details of a project.
    """
    return _private_sdk.get(project_id)  # type: ignore


_LIST_PROJECTS_EXAMPLE = """
from typing import Iterator

import anyscale
from anyscale.project.models import Project, ProjectSortField, ProjectSortOrder

projects: Iterator[Project] = anyscale.project.list(
    name_contains="my-project",
    creator_id="my-user-id",
    parent_cloud_id="my-cloud-id",
    include_defaults=True,
    max_items=20,
    page_size=10,
    sort_field=ProjectSortField.NAME,
    sort_order=ProjectSortOrder.ASC,
)
for project in projects:
    print(project.name)
"""

_LIST_PROJECTS_DOCSTRINGS = {
    "name_contains": "A string to filter projects by name.",
    "creator_id": "The ID of a creator to filter projects.",
    "parent_cloud_id": "The ID of a parent cloud to filter projects.",
    "include_defaults": "Whether to include default projects.",
    "max_items": "The maximum number of projects to return.",
    "page_size": "The number of projects to return per page.",
    "sort_field": "The field to sort projects by.",
    "sort_order": "The order to sort projects by.",
}


@sdk_command(
    _PROJECT_SDK_SINGLETON_KEY,
    PrivateProjectSDK,
    doc_py_example=_LIST_PROJECTS_EXAMPLE,
    arg_docstrings=_LIST_PROJECTS_DOCSTRINGS,
)
def list(  # noqa: A001
    *,
    name_contains: Optional[str] = None,
    creator_id: Optional[str] = None,
    parent_cloud_id: Optional[str] = None,
    include_defaults: bool = True,
    max_items: Optional[int] = None,
    page_size: Optional[int] = None,
    sort_field: Optional[ProjectSortField] = None,
    sort_order: Optional[ProjectSortOrder] = None,
    _private_sdk: Optional[PrivateProjectSDK] = None
) -> ResultIterator[Project]:
    """List projects.
    """
    return _private_sdk.list(  # type: ignore
        name_contains=name_contains,
        creator_id=creator_id,
        parent_cloud_id=parent_cloud_id,
        include_defaults=include_defaults,
        max_items=max_items,
        page_size=page_size,
        sort_field=sort_field,
        sort_order=sort_order,
    )


_CREATE_PROJECT_EXAMPLE = """
import anyscale

project_id: str = anyscale.project.create(
    name="my-project",
    parent_cloud_id="my-cloud-id",
    description="my-project-description",
)
"""

_CREATE_PROJECT_DOCSTRINGS = {
    "name": "The name of the project.",
    "parent_cloud_id": "The parent cloud that the project belongs to.",
    "description": "The description of the project.",
    "initial_cluster_config": "A YAML string containing the initial cluster config for the project.",
}


@sdk_command(
    _PROJECT_SDK_SINGLETON_KEY,
    PrivateProjectSDK,
    doc_py_example=_CREATE_PROJECT_EXAMPLE,
    arg_docstrings=_CREATE_PROJECT_DOCSTRINGS,
)
def create(
    name: str,
    parent_cloud_id: str,
    *,
    description: Optional[str] = None,
    initial_cluster_config: Optional[str] = None,
    _private_sdk: Optional[PrivateProjectSDK] = None
) -> str:
    """Create a project.
    """
    return _private_sdk.create(  # type: ignore
        name,
        description or "",
        parent_cloud_id,
        initial_cluster_config=initial_cluster_config,
    )


_DELETE_PROJECT_EXAMPLE = """
import anyscale

anyscale.project.delete(project_id="my-project-id")
"""

_DELETE_PROJECT_DOCSTRINGS = {
    "project_id": "The ID of the project to delete.",
}


@sdk_command(
    _PROJECT_SDK_SINGLETON_KEY,
    PrivateProjectSDK,
    doc_py_example=_DELETE_PROJECT_EXAMPLE,
    arg_docstrings=_DELETE_PROJECT_DOCSTRINGS,
)
def delete(project_id: str, *, _private_sdk: Optional[PrivateProjectSDK] = None):
    """Delete a project.
    """
    _private_sdk.delete(project_id)  # type: ignore


_GET_DEFAULT_PROJECT_EXAMPLE = """
import anyscale
from anyscale.project.models import Project

project: Project = anyscale.project.get_default(parent_cloud_id="my-cloud-id")
"""

_GET_DEFAULT_PROJECT_DOCSTRINGS = {
    "parent_cloud_id": "The ID of the parent cloud to get the default project for.",
}


@sdk_command(
    _PROJECT_SDK_SINGLETON_KEY,
    PrivateProjectSDK,
    doc_py_example=_GET_DEFAULT_PROJECT_EXAMPLE,
    arg_docstrings=_GET_DEFAULT_PROJECT_DOCSTRINGS,
)
def get_default(
    parent_cloud_id: str, *, _private_sdk: Optional[PrivateProjectSDK] = None
) -> Project:
    """Get the default project for a cloud.
    """
    return _private_sdk.get_default(parent_cloud_id)  # type: ignore



from typing import List

from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.client.openapi_client.models import Binding, UpdatePolicyRequest
from anyscale.policy.models import (
    Policy,
    PolicyBinding,
    PolicyConfig,
    PolicySyncStatus,
    ResourcePolicy,
)


class PrivatePolicySDK(BaseSDK):
    """Private SDK for resource policy operations."""

    def set(
        self, resource_type: str, resource_id: str, config: PolicyConfig,
    ):
        """
        Set user group permission policy for a resource.

        Args:
            resource_type: Resource type ('cloud', 'project', or 'organization')
            resource_id: Resource ID (e.g., cld_abc123, prj_xyz789)
            config: Policy configuration with role bindings
        """
        api_bindings = []
        for b in config.bindings:
            role = b.role_name
            if role == "collaborator" and resource_type in ("cloud", "project"):
                role = "write"
            api_bindings.append(Binding(role_name=role, principals=b.principals))
        api_policy = UpdatePolicyRequest(bindings=api_bindings)

        self.client.update_resource_policy(
            resource_type=resource_type, resource_id=resource_id, policy=api_policy,
        )

    def get(self, resource_type: str, resource_id: str,) -> Policy:
        """
        Get user group permission policy for a resource.

        Args:
            resource_type: Resource type ('cloud', 'project', or 'organization')
            resource_id: Resource ID (e.g., cld_abc123, prj_xyz789)

        Returns:
            Policy object with role bindings and sync status.
        """
        response = self.client.get_resource_policy(
            resource_type=resource_type, resource_id=resource_id,
        )

        bindings = []
        for b in response.bindings:
            role = b.role_name
            if role == "write" and resource_type in ("cloud", "project"):
                role = "collaborator"
            bindings.append(PolicyBinding(role_name=role, principals=b.principals))

        return Policy(
            bindings=bindings, sync_status=PolicySyncStatus(response.sync_status),
        )

    def list(self, resource_type: str,) -> List[ResourcePolicy]:
        """
        List permission policies for all resources of a specific type.

        Args:
            resource_type: Resource type to list policies for ('cloud' or 'project')

        Returns:
            List of ResourcePolicy objects with sync status.
        """
        # Convert singular to plural for API (API expects 'clouds' or 'projects')
        api_resource_type = f"{resource_type}s"
        response = self.client.list_resource_policies(resource_type=api_resource_type)

        policies = []
        for item in response.results:
            bindings = []
            for b in item.bindings:
                role = b.role_name
                if role == "write" and resource_type in ("cloud", "project"):
                    role = "collaborator"
                bindings.append(PolicyBinding(role_name=role, principals=b.principals))

            policies.append(
                ResourcePolicy(
                    resource_id=item.resource_id,
                    resource_type=item.resource_type,
                    bindings=bindings,
                    sync_status=PolicySyncStatus(item.sync_status),
                )
            )

        return policies



from contextlib import suppress
import os
from types import SimpleNamespace
from typing import cast, ClassVar, Dict, List, Optional

from anyscale._private.anyscale_client.common import AnyscaleClientInterface
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.decorated_application_template import (
    DecoratedApplicationTemplate,
)
from anyscale.image.models import ImageBuild, ImageBuildStatus
from anyscale.sdk.anyscale_client import (
    ClusterEnvironmentBuild,
    ClusterEnvironmentBuildStatus,
)


ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS = (
    os.environ.get("ANYSCALE_ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS", "0") == "1"
)


MAX_PAGE_SIZE = 50


class PrivateImageSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
    ):
        super().__init__(logger=logger, client=client)
        self._enable_image_build_for_tracked_requirements = (
            ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS
        )

    @property
    def enable_image_build_for_tracked_requirements(self) -> bool:
        return self._enable_image_build_for_tracked_requirements

    def get_default_image(self) -> str:
        return self.client.get_default_build_id()

    def get_image_build(self, build_id: str) -> Optional[ClusterEnvironmentBuild]:
        return self.client.get_cluster_env_build(build_id)

    _BACKEND_IMAGE_STATUS_TO_IMAGE_BUILD_STATUS: ClassVar[
        Dict[ClusterEnvironmentBuildStatus, ImageBuildStatus]
    ] = {
        ClusterEnvironmentBuildStatus.PENDING: ImageBuildStatus.IN_PROGRESS,
        ClusterEnvironmentBuildStatus.IN_PROGRESS: ImageBuildStatus.IN_PROGRESS,
        ClusterEnvironmentBuildStatus.SUCCEEDED: ImageBuildStatus.SUCCEEDED,
        ClusterEnvironmentBuildStatus.FAILED: ImageBuildStatus.FAILED,
        ClusterEnvironmentBuildStatus.PENDING_CANCELLATION: ImageBuildStatus.FAILED,
        ClusterEnvironmentBuildStatus.CANCELED: ImageBuildStatus.FAILED,
    }

    def _get_image_build_status(
        self, build: ClusterEnvironmentBuild
    ) -> ImageBuildStatus:
        return cast(
            ImageBuildStatus,
            self._BACKEND_IMAGE_STATUS_TO_IMAGE_BUILD_STATUS.get(
                build.status, ImageBuildStatus.UNKNOWN
            ),
        )

    def get(self, name: str) -> ImageBuild:
        """Get an image by name with optional version (name:version format).

        If no version is specified, returns the latest build.
        If a version is specified, searches for that specific revision.
        """
        image_name, version_str = name.split(":", 1) if ":" in name else (name, None)
        version: Optional[int] = None
        if version_str:
            try:
                version = int(version_str)
            except ValueError:
                raise ValueError(
                    f"Invalid version format '{version_str}'. Version must be a number."
                )

        response = self.client.list_application_templates(
            name=image_name,
            image_name=None,
            creator_id=None,
            project=None,
            include_archived=False,
            defaults_first=False,
            count=50,
            paging_token=None,
        )

        # list_application_templates does substring matching, need exact match
        results = response.results if response.results else []
        template = next((t for t in results if t.name == image_name), None)
        if not template:
            raise ValueError(f"Image '{image_name}' not found.")

        if not template.latest_build:
            raise ValueError(f"No builds found for image '{image_name}'.")

        # If specific version requested, verify it exists and get its details
        if version is not None and template.latest_build.revision != version:
            for b in self.client.list_cluster_env_builds(template.id):
                if b.revision == version and b.id:
                    image_uri_obj = self.get_image_uri_from_build_id(
                        b.id, use_image_alias=True
                    )
                    return ImageBuild(
                        id=template.id or "",
                        name=template.name or "",
                        project_id=template.project_id,
                        creator_id=template.creator_id,
                        creator_email=template.creator.email
                        if template.creator
                        else None,
                        is_anonymous=bool(template.anonymous),
                        created_at=template.created_at,
                        last_modified_at=template.last_modified_at,
                        latest_build_id=b.id,
                        latest_build_revision=b.revision,
                        latest_build_status=self._get_image_build_status(b),
                        latest_image_uri=image_uri_obj.image_uri
                        if image_uri_obj
                        else None,
                    )
            raise ValueError(f"Version {version} not found for image '{image_name}'.")

        return self._application_template_to_image_build(template)

    def build_image_from_containerfile(
        self,
        name: str,
        containerfile: str,
        ray_version: Optional[str] = None,
        anonymous: bool = False,
    ) -> str:
        return self.client.get_cluster_env_build_id_from_containerfile(
            cluster_env_name=name,
            containerfile=containerfile,
            anonymous=anonymous,
            ray_version=ray_version,
        )

    def build_image_from_containerfile_with_image_uri(
        self,
        name: str,
        containerfile: str,
        ray_version: Optional[str] = None,
        anonymous: bool = False,
    ) -> str:
        build_id = self.build_image_from_containerfile(
            name=name,
            containerfile=containerfile,
            ray_version=ray_version,
            anonymous=anonymous,
        )
        image_uri = self.get_image_uri_from_build_id(build_id)
        if image_uri:
            return image_uri.image_uri
        raise RuntimeError(
            f"This is a bug! Failed to get image uri for build {build_id} that just created."
        )

    def build_image_from_requirements(
        self, name: str, base_build_id: str, requirements: List[str]
    ):
        if requirements:
            base_build = self.client.get_cluster_env_build(base_build_id)
            if (
                base_build
                and base_build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
                and base_build.docker_image_name
            ):
                self.logger.info(f"Using tracked python packages: {requirements}")
                lines = [
                    "# syntax=docker/dockerfile:1",
                    f"FROM {base_build.docker_image_name}",
                ]
                for requirement in requirements:
                    lines.append(f'RUN pip install "{requirement}"')
                return self.build_image_from_containerfile(
                    name, "\n".join(lines), ray_version=base_build.ray_version
                )
            else:
                raise RuntimeError(
                    f"Base build {base_build_id} is not a successful build."
                )
        else:
            return base_build_id

    def registery_image(
        self,
        image_uri: str,
        registry_login_secret: Optional[str] = None,
        ray_version: Optional[str] = None,
    ) -> str:
        image_uri_checked = ImageURI.from_str(image_uri_str=image_uri)
        return self.client.get_cluster_env_build_id_from_image_uri(
            image_uri=image_uri_checked,
            registry_login_secret=registry_login_secret,
            ray_version=ray_version,
        )

    def register_byod_image_with_name(
        self,
        image_uri_str: str,
        name: str,
        ray_version: Optional[str] = None,
        registry_login_secret: Optional[str] = None,
    ) -> str:
        image_uri_checked = ImageURI.from_str(image_uri_str=image_uri_str)
        if image_uri_checked.is_cluster_env_image():
            raise RuntimeError(
                f"Image URI {image_uri_str} is not a BYOD image. "
                "The 'register' command only works with BYOD images."
            )

        build_id = self.client.get_cluster_env_build_id_from_image_uri(
            image_uri=image_uri_checked,
            registry_login_secret=registry_login_secret,
            ray_version=ray_version,
            name=name,
        )

        fetched_image_uri = self.get_image_uri_from_build_id(
            build_id, use_image_alias=True
        )
        if fetched_image_uri:
            return fetched_image_uri.image_uri
        raise RuntimeError(
            f"This is a bug! Failed to get image uri for build {build_id} that just created."
        )

    def get_image_uri_from_build_id(
        self, build_id: str, use_image_alias: bool = False
    ) -> Optional[ImageURI]:
        return self.client.get_cluster_env_build_image_uri(
            cluster_env_build_id=build_id, use_image_alias=use_image_alias,
        )

    def _application_template_to_image_build(
        self, template: DecoratedApplicationTemplate
    ) -> ImageBuild:
        """Convert a DecoratedApplicationTemplate to an ImageBuild."""
        latest_build = template.latest_build
        latest_build_id = latest_build.id if latest_build else None
        latest_build_revision = latest_build.revision if latest_build else None
        latest_build_status = None
        latest_image_uri = None

        if latest_build:
            if latest_build.status:
                with suppress(ValueError):
                    latest_build_status = ImageBuildStatus.validate(
                        latest_build.status.upper()
                    )

            latest_image_uri = latest_build.docker_image_name
            if not latest_image_uri and latest_build_id:
                image_uri = self.get_image_uri_from_build_id(
                    latest_build_id, use_image_alias=True
                )
                if image_uri:
                    latest_image_uri = image_uri.image_uri

        creator_email = template.creator.email if template.creator else None

        return ImageBuild(
            id=template.id or "",
            name=template.name or "",
            project_id=template.project_id,
            creator_id=template.creator_id,
            creator_email=creator_email,
            is_anonymous=bool(template.anonymous),
            created_at=template.created_at,
            last_modified_at=template.last_modified_at,
            latest_build_id=latest_build_id,
            latest_build_revision=latest_build_revision,
            latest_build_status=latest_build_status,
            latest_image_uri=latest_image_uri,
        )

    def _should_filter_template(
        self,
        template: DecoratedApplicationTemplate,
        include_anonymous: bool,
        include_archived: bool,
    ) -> bool:
        """Return True if template should be filtered out."""
        if not include_anonymous and template.anonymous:
            return True
        return bool(not include_archived and template.archived_at is not None)

    def list(  # noqa: PLR0913, PLR0917
        self,
        *,
        image_id: Optional[str] = None,
        name: Optional[str] = None,
        image_name: Optional[str] = None,
        project: Optional[str] = None,
        creator_id: Optional[str] = None,
        include_archived: bool = False,
        include_anonymous: bool = False,
        max_items: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ResultIterator[ImageBuild]:
        """List images or fetch a single image by ID."""
        if page_size is not None and not (1 <= page_size <= MAX_PAGE_SIZE):
            raise ValueError(
                f"page_size must be between 1 and {MAX_PAGE_SIZE}, inclusive."
            )

        # Handle single image lookup by ID
        if image_id is not None:
            template = self.client.get_application_template(image_id)

            if template is None or self._should_filter_template(
                template, include_anonymous, include_archived
            ):
                # Return empty iterator
                return ResultIterator(
                    page_token=None,
                    max_items=0,
                    fetch_page=lambda _: SimpleNamespace(
                        results=[], metadata=SimpleNamespace(next_paging_token=None),
                    ),
                    parse_fn=None,
                )

            # Return single-item iterator
            summary = self._application_template_to_image_build(template)
            return ResultIterator(
                page_token=None,
                max_items=1,
                fetch_page=lambda token: SimpleNamespace(
                    results=[summary] if token is None else [],
                    metadata=SimpleNamespace(next_paging_token=None),
                ),
                parse_fn=None,
            )

        # Handle paginated list
        def _fetch_page(token: Optional[str]) -> SimpleNamespace:
            response = self.client.list_application_templates(
                name=name,
                image_name=image_name,
                creator_id=creator_id,
                project=project,
                include_archived=include_archived,
                defaults_first=False,
                count=page_size,
                paging_token=token,
            )

            results = response.results if response.results else []
            filtered_results = [
                template
                for template in results
                if not self._should_filter_template(
                    template, include_anonymous, include_archived
                )
            ]

            next_token = (
                response.metadata.next_paging_token if response.metadata else None
            )

            return SimpleNamespace(
                results=filtered_results,
                metadata=SimpleNamespace(next_paging_token=next_token),
            )

        return ResultIterator(
            page_token=None,
            max_items=max_items,
            fetch_page=_fetch_page,
            parse_fn=self._application_template_to_image_build,
        )

    def archive(
        self, *, name: Optional[str] = None, image_id: Optional[str] = None
    ) -> None:  # noqa: A002
        """Archive an image by name or ID.

        If name is provided, looks up the image ID first.
        Once archived, the image name will no longer be usable in the organization.

        Args:
            name: The name of the image to archive (can include version, e.g., 'name:version').
            image_id: The ID of the image to archive.

        Raises:
            ValueError: If the image is not found or cannot be archived.
        """
        if name is None and image_id is None:
            raise ValueError("Either 'name' or 'image_id' must be provided.")

        if name is not None and image_id is not None:
            raise ValueError(
                "Only one of 'name' or 'image_id' can be provided, not both."
            )

        resolved_id = image_id
        if name is not None:
            # Parse name to extract just the image name (without version)
            image_name = name.split(":", 1)[0] if ":" in name else name

            # Look up the image by name
            response = self.client.list_application_templates(
                name=image_name,
                image_name=None,
                creator_id=None,
                project=None,
                include_archived=False,
                defaults_first=False,
                count=50,
                paging_token=None,
            )

            results = response.results if response.results else []
            template = next((t for t in results if t.name == image_name), None)
            if not template:
                raise ValueError(f"Image '{image_name}' not found.")
            resolved_id = template.id

        self.client.archive_image(image_id=resolved_id)
        self.logger.info("Successfully archived image.")



from io import StringIO
from json import dumps as json_dumps
import pathlib
from typing import Any, Dict, List, Optional

import click
from rich.console import Console
from rich.table import Table
import tabulate
import yaml

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.list_util import (
    create_table,
    display_list,
    NON_INTERACTIVE_DEFAULT_MAX_ITEMS,
    validate_page_size,
)
from anyscale.commands.util import AnyscaleCommand
from anyscale.controllers.schedule_controller import ScheduleController
from anyscale.schedule.models import (
    JobConfig,
    ScheduleConfig,
    ScheduleState,
    ScheduleStatus,
)
from anyscale.util import get_endpoint, validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("schedule", help="Create and manage Anyscale Schedules.")
def schedule_cli() -> None:
    pass


def _read_identifiers_from_config_file(path: str):
    """Read the 'name', 'cloud', and 'project' properties from the config file at `path`.

    Return the identifers as a ScheduleIdentifiers object.
    """
    if not pathlib.Path(path).is_file():
        raise click.ClickException(f"Config file not found at path: '{path}'.")

    with open(path) as f:
        config = yaml.safe_load(f)

    if config is None or "job_config" not in config:
        raise click.ClickException(
            f"No 'job_config' property found in config file '{path}'."
        )

    job_config = config.get("job_config")
    name = job_config.get("name", None)
    cloud = job_config.get("cloud", None)
    project = job_config.get("project", None)

    return name, cloud, project


def _validate_schedule_identifiers(
    name: Optional[str], id: Optional[str], config_file: Optional[str]  # noqa: A002
):
    num_passed = sum(val is not None for val in [name, id, config_file])
    if num_passed == 0:
        raise click.ClickException(
            "One of '--name', '--id', or '--config-file' must be provided."
        )

    if num_passed > 1:
        raise click.ClickException(
            "Only one of '--name', '--id', and '--config-file' can be provided."
        )


@schedule_cli.command(
    name="apply", cls=AnyscaleCommand, example=command_examples.SCHEDULE_APPLY_EXAMPLE
)
@click.option(
    "--config-file",
    "-f",
    required=True,
    type=str,
    help="Path to a YAML config file to use for this schedule. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
def apply(config_file: str, name: Optional[str],) -> None:
    """ Create or Update a Schedule

    The schedule should be specified in a YAML config file.
    """
    if not pathlib.Path(config_file).is_file():
        raise click.ClickException(f"Schedule config file '{config_file}' not found.")

    config = ScheduleConfig.from_yaml(config_file)

    if name is not None:
        assert isinstance(config.job_config, JobConfig)
        config = config.options(job_config=config.job_config.options(name=name),)

    log.info(f"Applying schedule with config {config}.")
    anyscale.schedule.apply(config)


def _create_schedules_table_v2(is_first: bool) -> Table:
    """Create a Rich Table for displaying schedules in v2 mode.

    Args:
        is_first: Whether this is the first page (controls header display).

    Returns:
        Rich Table configured for schedule display.
    """
    columns = [
        ("ID", "cyan", True),
        ("Name", None, False),
        ("State", "green", False),
        ("Cron Expression", None, False),
        ("Timezone", None, False),
        ("Project", None, False),
    ]
    return create_table(columns, is_first)


def _format_schedule_row_v2(schedule: ScheduleStatus) -> Dict[str, Any]:
    """Format a ScheduleStatus for table row or JSON output.

    Args:
        schedule: The ScheduleStatus object to format.

    Returns:
        Dictionary with formatted schedule data.
    """
    return {
        "id": schedule.id or "",
        "name": schedule.name or "",
        "state": str(schedule.state) if schedule.state else "",
        "cron_expression": schedule.config.cron_expression if schedule.config else "",
        "timezone": schedule.config.timezone if schedule.config else "",
        "project": schedule.config.job_config.project
        if schedule.config and schedule.config.job_config
        else "",
    }


def _display_schedules_table(schedules: List[ScheduleStatus]) -> None:
    """Display schedules in a tabulated format."""
    if not schedules:
        print("No schedules found.")
        return

    schedules_table = [
        [
            schedule.name,
            schedule.id,
            str(schedule.state) if schedule.state else None,
            schedule.config.cron_expression if schedule.config else None,
            schedule.config.timezone if schedule.config else None,
        ]
        for schedule in schedules
    ]

    table = tabulate.tabulate(
        schedules_table,
        headers=["NAME", "ID", "STATE", "CRON", "TIMEZONE"],
        tablefmt="plain",
    )
    print(f"SCHEDULES:\n{table}")

    endpoint = get_endpoint("")
    print(f"\nView your schedules at: {endpoint}schedules")


def _print_schedule_list_diagnostics(  # noqa: PLR0913
    stderr: Console,
    name: Optional[str],
    schedule_id: Optional[str],
    project: Optional[str],
    cloud: Optional[str],
    creator_id: Optional[str],
    include_all_users: bool,
    interactive: bool,
    page_size: int,
    effective_max: Optional[int],
) -> None:
    """Prints diagnostic information for the list command."""
    stderr.print("[bold]Listing schedules with:[/]")
    stderr.print(f"• name            = {name or '<any>'}")
    stderr.print(f"• id              = {schedule_id or '<any>'}")
    stderr.print(f"• project         = {project or '<any>'}")
    stderr.print(f"• cloud           = {cloud or '<any>'}")
    stderr.print(f"• creator_id      = {creator_id or '<any>'}")
    stderr.print(f"• include_all     = {include_all_users}")
    stderr.print(f"• mode            = {'interactive' if interactive else 'batch'}")
    stderr.print(f"• per-page limit  = {page_size}")
    stderr.print(f"• max-items total = {effective_max if effective_max else 'all'}")
    stderr.print(f"\nView your Schedules in the UI at {get_endpoint('/schedules')}\n")


# TODO(praneethkaturi): Add --sort option for schedule list.
# Requires backend changes:
#   - backend/server/experimental_cron/cron_jobs_base_model.py (add ScheduleSortField enum)
#   - backend/server/experimental_cron/cron_jobs_router.py (add sort query params)
#   - backend/server/experimental_cron/cron_jobs_dao.py (dynamic sorting)
# And SDK changes:
#   - frontend/cli/anyscale/schedule/_private/schedule_sdk.py
#   - frontend/cli/anyscale/schedule/commands.py
@schedule_cli.command(
    name="list", cls=AnyscaleCommand, example=command_examples.SCHEDULE_LIST_EXAMPLE
)
@click.option(
    "--v2",
    is_flag=True,
    default=False,
    help="[RECOMMENDED] Enable extended filtering options. Needs migration to match return values.",
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Filter by the name of the schedule.",
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--project",
    required=False,
    default=None,
    help="The named Anyscale Project for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the project of the workspace). Only with --v2 flag.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or,if running in a workspace, the cloud of the workspace). Only with --v2 flag.",
)
@click.option(
    "--creator-id",
    required=False,
    default=None,
    help="Filter by creator ID. Only with --v2 flag.",
)
@click.option(
    "--max-items",
    type=int,
    callback=validate_non_negative_arg,
    help="Max total items (only with --no-interactive).",
)
@click.option(
    "--page-size",
    required=False,
    default=None,
    type=int,
    help="Number of items per page (1-50). Only with --v2.",
    callback=validate_page_size,
)
@click.option(
    "--json",
    "-j",
    "json_output",
    is_flag=True,
    default=False,
    help="Output results as JSON. Only with --v2.",
)
@click.option(
    "--interactive/--no-interactive",
    default=True,
    show_default=True,
    help="Enable interactive pagination. Only with --v2.",
)
@click.option(
    "--include-all-users/--only-mine",
    default=False,
    help="Include schedules from all users. Only with --v2.",
)
def list(  # noqa: A001 PLR0913
    v2: bool,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    project: Optional[str] = None,
    cloud: Optional[str] = None,
    creator_id: Optional[str] = None,
    max_items: Optional[int] = None,
    page_size: Optional[int] = None,
    json_output: bool = False,
    interactive: bool = True,
    include_all_users: bool = False,
) -> None:
    """List Schedules

    You can optionally filter schedules by name, project, cloud, or creator.
    """
    if v2:
        # Validate max_items only allowed with --no-interactive (v2 only)
        if max_items is not None and interactive:
            raise click.UsageError("--max-items only allowed with --no-interactive")
        # New SDK path with pagination and output options

        # Apply defaults for v2-only options
        effective_page_size = page_size if page_size is not None else 10

        # Compute effective max_items for non-interactive mode
        effective_max = max_items
        if not interactive and effective_max is None:
            effective_max = NON_INTERACTIVE_DEFAULT_MAX_ITEMS

        # Print diagnostics header (not in JSON mode)
        if not json_output:
            stderr = Console(stderr=True)
            _print_schedule_list_diagnostics(
                stderr=stderr,
                name=name,
                schedule_id=id,
                project=project,
                cloud=cloud,
                creator_id=creator_id,
                include_all_users=include_all_users,
                interactive=interactive,
                page_size=effective_page_size,
                effective_max=effective_max if not interactive else None,
            )

        iterator = anyscale.schedule.list(
            name=name,
            schedule_id=id,
            project=project,
            cloud=cloud,
            creator_id=creator_id,
            include_all_users=include_all_users,
            page_size=effective_page_size,
            max_items=effective_max if not interactive else None,
        )

        console = Console()
        total = display_list(
            iterator=iterator,
            item_formatter=_format_schedule_row_v2,
            table_creator=_create_schedules_table_v2,
            json_output=json_output,
            page_size=effective_page_size,
            interactive=interactive,
            max_items=effective_max if not interactive else None,
            console=console,
        )
        if not json_output:
            if total:
                stderr.print(f"\nFetched {total} schedule(s).")
            else:
                stderr.print("\nNo schedules found.")
    else:
        # Legacy path with deprecation warning
        # Check if v2-only options are being used without --v2
        if any(
            [
                project,
                cloud,
                creator_id,
                max_items is not None,
                page_size is not None,
                json_output,
                include_all_users,
                not interactive,
            ]
        ):
            click.echo(
                "ERROR: Options --project, --cloud, --creator-id, --max-items, "
                "--page-size, --json, --include-all-users, and --no-interactive require --v2 flag.\n"
                "Use: anyscale schedule list --v2 [options]",
                err=True,
            )
            raise click.exceptions.Exit(1)

        job_controller = ScheduleController()
        job_controller.list(name=name, id=id)


@schedule_cli.command(
    name="pause", cls=AnyscaleCommand, example=command_examples.SCHEDULE_PAUSE_EXAMPLE
)
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def pause(
    config_file: str, name: str, cloud: str, project: str, id: str  # noqa: A002
) -> None:
    """Pause a Schedule.

    You can pause a schedule by config file, name, or id.

    To specify the schedule by name, use the --name flag. You can specify the cloud with --cloud and the project with --project.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by config file, use --config-file. Ensure that name and optionally cloud and project are specified in the
    config file's job config.
    """
    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        anyscale.schedule.set_state(id=id, state=ScheduleState.DISABLED)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        anyscale.schedule.set_state(
            name=name, cloud=cloud, project=project, state=ScheduleState.DISABLED
        )


@schedule_cli.command(
    name="resume", cls=AnyscaleCommand, example=command_examples.SCHEDULE_RESUME_EXAMPLE
)
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def resume(
    config_file: str, name: str, cloud: str, project: str, id: str  # noqa: A002
) -> None:
    """ Resume a Schedule

    You can resume a schedule by config file, name, or id.

    To specify the schedule by name, use the --name flag. You can specify the cloud with --cloud and the project with --project.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by config file, use --config-file. Ensure that name and optionally cloud and project are specified in the
    config file's job config.
    """
    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        anyscale.schedule.set_state(id=id, state=ScheduleState.ENABLED)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        anyscale.schedule.set_state(
            name=name, cloud=cloud, project=project, state=ScheduleState.ENABLED
        )


@schedule_cli.command(
    name="status", cls=AnyscaleCommand, example=command_examples.SCHEDULE_STATUS_EXAMPLE
)
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--json",
    "-j",
    is_flag=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Include verbose details in the status.",
)
def status(
    config_file: str,
    name: str,
    cloud: str,
    project: str,
    id: str,  # noqa: A002
    json: bool,
    verbose: bool,
) -> None:
    """Query the status of a Schedule.

    You can query the status of a schedule by config file, name, or id.

    To specify the schedule by name, use the --name flag. You can specify the cloud with --cloud and the project with --project.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by config file, use --config-file. Ensure that name and optionally cloud and project are specified in the
    config file's job config.
    """
    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        status = anyscale.schedule.status(id=id)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        status = anyscale.schedule.status(name=name, cloud=cloud, project=project)

    status_dict = status.to_dict()

    if not verbose:
        status_dict.pop("config", None)

    if json:
        print(json_dumps(status_dict, indent=4, sort_keys=False))
    else:
        stream = StringIO()
        yaml.dump(status_dict, stream, sort_keys=False)
        print(stream.getvalue(), end="")


@schedule_cli.command(
    name="run", cls=AnyscaleCommand, example=command_examples.SCHEDULE_RUN_EXAMPLE
)
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def trigger(
    config_file: str, name: str, id: str, cloud: str, project: str  # noqa: A002
) -> None:
    """ Manually run a Schedule

    This function takes an existing schedule and runs it now.
    You can specify the schedule by name or id.
    You can also pass in a YAML file as a convinience. This is equivalent to passing in the name specified in the YAML file.
    IMPORTANT: if you pass in a YAML definition that differs from the Schedule defition, the Schedule will NOT be updated.
    Please use the `anyscale schedule apply` command to update the configuration of your schedule
    or use the `anyscale job submit` command to submit a one off job that is not a part of a schedule.
    """

    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        anyscale.schedule.trigger(id=id)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        anyscale.schedule.trigger(
            name=name, cloud=cloud, project=project,
        )


@schedule_cli.command(
    name="url", cls=AnyscaleCommand, example=command_examples.SCHEDULE_URL_EXAMPLE
)
@click.argument("schedule_config_file", required=False)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option("--v2", is_flag=True, help="Use new SDK-based implementation")
@click.option("--cloud", help="Cloud name (required with --name in v2 mode)")
@click.option("--project", help="Project name (required with --name in v2 mode)")
def url(
    schedule_config_file: str,
    id: str,  # noqa: A002
    name: str,
    v2: bool,
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Get a Schedule URL

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    You can also specify the schedule by name or id.
    """
    if v2:
        result_url = anyscale.schedule.url(
            id=id, name=name, cloud=cloud, project=project,
        )
        click.echo(f"View your schedule at {result_url}")
    else:
        job_controller = ScheduleController()
        resolved_id = job_controller.resolve_file_name_or_id(
            schedule_config_file=schedule_config_file, id=id, name=name
        )
        job_controller.url(resolved_id)


def _validate_delete_identifiers(
    name: Optional[str],
    id: Optional[str],  # noqa: A002
    cloud: Optional[str],
    project: Optional[str],
):
    """Validate identifiers for the delete command.

    Either --id OR --name must be provided.
    When --name is used, --cloud and --project are also required.
    When --id is used, --cloud and --project cannot be used.
    """
    if name is None and id is None:
        raise click.ClickException("One of '--name' or '--id' must be provided.")

    if name is not None and id is not None:
        raise click.ClickException("Only one of '--name' or '--id' can be provided.")

    if id is not None and (cloud is not None or project is not None):
        raise click.ClickException(
            "'--cloud' and '--project' cannot be used with '--id'."
        )

    if name is not None and (cloud is None or project is None):
        raise click.ClickException(
            "'--cloud' and '--project' are required when using '--name'."
        )


@schedule_cli.command(
    name="delete", cls=AnyscaleCommand, example=command_examples.SCHEDULE_DELETE_EXAMPLE
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule (required with --name).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project for the schedule (required with --name).",
)
def delete(
    name: Optional[str],
    id: Optional[str],  # noqa: A002
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Delete a Schedule.

    If the schedule is active, it will be automatically paused before deletion.
    The schedule must have no active triggered jobs.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by name, use the --name flag along with --cloud and --project.
    """
    _validate_delete_identifiers(name=name, id=id, cloud=cloud, project=project)

    if id is not None:
        anyscale.schedule.delete(id=id)
    else:
        anyscale.schedule.delete(name=name, cloud=cloud, project=project)



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, ClassVar, DefaultDict, Dict, List, Optional, Union
import warnings

from anyscale._private.models import ModelBase, ModelEnum


ResourceDict = Dict[str, float]
LabelDict = Dict[str, str]
AdvancedInstanceConfigDict = Dict[str, Any]

# CPU architecture constants
CPU_ARCHITECTURE_X86_64 = "x86_64"
CPU_ARCHITECTURE_ARM64 = "arm64"

# TPU configuration constants
# See: https://cloud.google.com/kubernetes-engine/docs/concepts/plan-tpus
TPU_DOCS_URL = "https://cloud.google.com/kubernetes-engine/docs/concepts/plan-tpus"
REQUIRED_TPU_NODE_SELECTORS = [
    "cloud.google.com/gke-tpu-topology",
    "cloud.google.com/gke-tpu-accelerator",
]
# Ray labels that can be used to derive TPU node selectors (alternative to explicit node selectors)
RAY_TPU_ACCELERATOR_LABEL = "ray.io/accelerator-type"
RAY_TPU_TOPOLOGY_LABEL = "ray.io/tpu-topology"


def _extract_accelerator_from_labels(
    labels: Optional[LabelDict], required_labels: Optional[LabelDict]
) -> Optional[str]:
    """Extract accelerator type from labels or required_labels.

    Checks for the ray.io/accelerator-type label in required_labels first,
    then falls back to labels. This allows users to specify GPU/TPU type
    via labels as an alternative to the accelerator field in required_resources.

    Args:
        labels: Optional labels dictionary
        required_labels: Optional required_labels dictionary

    Returns:
        The accelerator type string if found, None otherwise.
    """
    # Check required_labels first (takes precedence)
    if required_labels and RAY_TPU_ACCELERATOR_LABEL in required_labels:
        accel = required_labels[RAY_TPU_ACCELERATOR_LABEL]
        if accel:
            return accel

    # Fall back to labels
    if labels and RAY_TPU_ACCELERATOR_LABEL in labels:
        accel = labels[RAY_TPU_ACCELERATOR_LABEL]
        if accel:
            return accel

    return None


def _generate_free_pod_shape_name(
    cpu: Optional[int],
    memory: Optional[Union[str, int]],
    gpu: Optional[int],
    accelerator: Optional[str],
    tpu: Optional[int],
    tpu_hosts: Optional[int],
    cpu_architecture: Optional[str] = None,
) -> str:
    """Generate a unique name for a free pod shape based on physical resources.

    Follows the backend naming convention:
    - "FP-{CPU}CPU-{MEM}GB" for CPU/memory only
    - "FP-{CPU}CPU-{MEM}GB-{GPU}GPU" for GPU without accelerator type
    - "FP-{CPU}CPU-{MEM}GB-{GPU}GPU-{ACCEL}" for GPU with accelerator type
    - "FP-{CPU}CPU-{MEM}GB-{TPU}TPU-{TPUHOST}TPUHost-{ACCEL}" for TPU
    - Append "-ARM64" for arm64 architecture

    Args:
        cpu: Number of CPUs
        memory: Memory in bytes or Kubernetes format string (e.g., '4Gi')
        gpu: Number of GPUs
        accelerator: Accelerator type (e.g., 'T4', 'A100', 'TPU-V6E')
        tpu: Number of TPUs
        tpu_hosts: Number of TPU hosts
        cpu_architecture: CPU architecture type ('x86_64' or 'arm64')

    Returns:
        Generated free pod shape name.
    """
    # Parse memory to bytes if string, then convert to GB
    memory_bytes = _parse_memory_string(memory) if memory else 0
    memory_gb = memory_bytes // (1024 * 1024 * 1024)  # Convert to GB

    parts = [f"FP-{cpu or 0}CPU-{memory_gb}GB"]

    if tpu and tpu > 0:
        parts.append(f"{tpu}TPU")
        if tpu_hosts and tpu_hosts > 0:
            parts.append(f"{tpu_hosts}TPUHost")
        if accelerator:
            parts.append(accelerator)
    elif gpu and gpu > 0:
        parts.append(f"{gpu}GPU")
        if accelerator:
            parts.append(accelerator)

    # Add ARM64 suffix for non-default architecture
    if cpu_architecture and cpu_architecture.lower() == CPU_ARCHITECTURE_ARM64:
        parts.append("ARM64")

    return "-".join(parts)


def _parse_memory_string(memory_str: Union[str, int]) -> int:
    """Parse memory string (e.g., '4Gi', '1024Mi', '1G') to bytes.

    Uses Kubernetes quantity parsing to handle all standard K8s resource formats.

    Args:
        memory_str: Memory value as string (e.g., '4Gi') or int (bytes).

    Returns:
        Memory in bytes as int.
    """
    if isinstance(memory_str, int):
        return memory_str

    if not isinstance(memory_str, str):
        raise TypeError(f"Memory must be a string or int, got: {type(memory_str)}")

    try:
        from kubernetes.utils.quantity import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional Kubernetes dependency for quantity parsing")
            parse_quantity,
        )

        # parse_quantity returns a Decimal in the canonical unit (bytes for memory)
        return int(parse_quantity(memory_str))
    except (ValueError, ImportError, AttributeError) as e:
        raise ValueError(f"Invalid memory format: {memory_str}. Error: {e}") from e


def _format_bytes_to_memory_string(memory_bytes: Union[int, str]) -> str:
    """Convert memory in bytes to human-readable string format (e.g., '8Gi').

    Args:
        memory_bytes: Memory value in bytes (int) or already formatted string.

    Returns:
        Memory as human-readable string (e.g., '8Gi', '512Mi').
    """
    if isinstance(memory_bytes, str):
        return memory_bytes

    if not isinstance(memory_bytes, int):
        return str(memory_bytes)

    # Use GiB if it's a clean multiple, otherwise use MiB or bytes
    gib = 1024 * 1024 * 1024
    mib = 1024 * 1024

    # GiB format (clean multiple or with decimal)
    if memory_bytes >= gib:
        if memory_bytes % gib == 0:
            return f"{memory_bytes // gib}Gi"
        return f"{memory_bytes / gib:.1f}Gi".rstrip("0").rstrip(".")

    # MiB format
    if memory_bytes >= mib:
        return f"{memory_bytes // mib}Mi"

    return str(memory_bytes)


@dataclass(frozen=True)
class PhysicalResources(ModelBase):
    """Physical resources specification for compute nodes.

    Used for custom instance types (free pod shapes) to explicitly define
    CPU, memory, and GPU resources instead of deriving from instance type.
    """

    __doc_py_example__ = """
from anyscale.compute_config.models import PhysicalResources

required_resources = PhysicalResources(
    CPU=4,
    memory="8Gi",
    GPU=1,
    accelerator="T4",
)
"""

    __doc_yaml_example__ = """
required_resources:
  CPU: 4
  memory: 8Gi
  GPU: 1
  accelerator: T4
"""

    CPU: Optional[int] = field(
        default=None, metadata={"docstring": "Number of CPUs to allocate."},
    )

    def _validate_CPU(self, CPU: Optional[int]):
        if CPU is not None:
            if not isinstance(CPU, int):
                raise TypeError("'CPU' must be an int.")
            if CPU < 0:
                raise ValueError("'CPU' must be >= 0.")

    memory: Optional[Union[str, int]] = field(
        default=None,
        metadata={
            "docstring": "Amount of memory to allocate. Can be specified as bytes (int) or as a string with units (e.g., '4Gi', '1024Mi')."
        },
    )

    def _validate_memory(self, memory: Optional[Union[str, int]]):
        if memory is not None:
            try:
                parsed = _parse_memory_string(memory)
                if parsed < 0:
                    raise ValueError("'memory' must be >= 0.")
            except TypeError:
                # Re-raise TypeError as-is for better error messages
                raise
            except ValueError as e:
                raise ValueError(f"Invalid memory value: {e}")

    GPU: Optional[int] = field(
        default=None, metadata={"docstring": "Number of GPUs to allocate."},
    )

    def _validate_GPU(self, GPU: Optional[int]):
        if GPU is not None:
            if not isinstance(GPU, int):
                raise TypeError("'GPU' must be an int.")
            if GPU < 0:
                raise ValueError("'GPU' must be >= 0.")

    accelerator: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Type of accelerator (e.g., 'T4', 'L4', 'A100', 'H100', 'TPUv4'). "
            "Use abbreviated names for readability in generated instance type names."
        },
    )

    def _validate_accelerator(self, accelerator: Optional[str]):
        if accelerator is not None and not isinstance(accelerator, str):
            raise TypeError("'accelerator' must be a string.")

    TPU: Optional[int] = field(
        default=None, metadata={"docstring": "Number of TPUs to allocate."},
    )

    def _validate_TPU(self, TPU: Optional[int]):
        if TPU is not None:
            if not isinstance(TPU, int):
                raise TypeError("'TPU' must be an int.")
            if TPU < 0:
                raise ValueError("'TPU' must be >= 0.")

    tpu_hosts: Optional[int] = field(
        default=None,
        metadata={
            "docstring": "Number of TPU hosts (for anyscale/tpu_hosts custom resource).",
            "alias": "anyscale/tpu_hosts",
        },
    )

    def _validate_tpu_hosts(self, tpu_hosts: Optional[int]):
        if tpu_hosts is not None:
            if not isinstance(tpu_hosts, int):
                raise TypeError("'tpu_hosts' must be an int.")
            if tpu_hosts < 0:
                raise ValueError("'tpu_hosts' must be >= 0.")

    cpu_architecture: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "CPU architecture type. Valid values: 'x86_64' (default), 'arm64'.",
        },
    )

    def _validate_cpu_architecture(self, cpu_architecture: Optional[str]):
        if cpu_architecture is not None:
            if not isinstance(cpu_architecture, str):
                raise TypeError("'cpu_architecture' must be a string.")
            valid_values = [CPU_ARCHITECTURE_X86_64, CPU_ARCHITECTURE_ARM64]
            if cpu_architecture not in valid_values:
                raise ValueError(
                    f"'cpu_architecture' must be one of {valid_values}, got '{cpu_architecture}'."
                )

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "PhysicalResources":
        """Create PhysicalResources from dict, handling case variations.

        Handles conversions from API model (lowercase) to user model (mixed case):
        - cpu -> CPU
        - gpu -> GPU
        - tpu -> TPU
        - Memory -> memory
        - anyscale_tpu_hosts -> tpu_hosts

        Also converts memory from bytes (int) to human-readable format (e.g., '8Gi').
        """
        # Key mapping from API model (lowercase) to user model (mixed case)
        key_mapping = {
            "cpu": "CPU",
            "gpu": "GPU",
            "tpu": "TPU",
            "Memory": "memory",
            "anyscale_tpu_hosts": "tpu_hosts",
        }

        normalized = {}
        for k, v in d.items():
            # Skip None values
            if v is None:
                continue
            # Map to expected key name, or keep original if not in mapping
            normalized_key = key_mapping.get(k, k)
            # Convert memory from bytes to human-readable format
            normalized_value = (
                _format_bytes_to_memory_string(v)
                if normalized_key == "memory" and isinstance(v, int)
                else v
            )
            normalized[normalized_key] = normalized_value

        return cls(**normalized)

    def to_dict(
        self, *, exclude_none: bool = True, for_api: bool = False
    ) -> Dict[str, Any]:
        """Convert to dictionary.

        Args:
            exclude_none: If True, exclude keys with None values.
            for_api: If True, use lowercase keys (cpu, gpu, tpu) and convert memory to bytes.
                     If False, use uppercase keys (CPU, GPU, TPU) and keep memory human-readable.
        """
        result: Dict[str, Any] = {}

        # Key names based on output format
        cpu_key = "cpu" if for_api else "CPU"
        gpu_key = "gpu" if for_api else "GPU"
        tpu_key = "tpu" if for_api else "TPU"

        if self.CPU is not None or not exclude_none:
            result[cpu_key] = self.CPU

        if self.memory is not None:
            if for_api:
                # Convert to bytes for API
                result["memory"] = _parse_memory_string(self.memory)
            # Keep human-readable format for user output
            elif isinstance(self.memory, str):
                result["memory"] = self.memory
            else:
                result["memory"] = _format_bytes_to_memory_string(self.memory)
        elif not exclude_none:
            result["memory"] = None

        if self.GPU is not None or not exclude_none:
            result[gpu_key] = self.GPU

        if self.accelerator is not None or not exclude_none:
            result["accelerator"] = self.accelerator

        if self.TPU is not None or not exclude_none:
            result[tpu_key] = self.TPU

        if self.tpu_hosts is not None or not exclude_none:
            result["tpu_hosts"] = self.tpu_hosts

        if self.cpu_architecture is not None or not exclude_none:
            result["cpu_architecture"] = self.cpu_architecture

        return result


def _validate_resource_dict(r: Optional[ResourceDict], *, field_name: str):
    if r is None:
        return

    if not isinstance(r, dict):
        raise TypeError(f"'{field_name}' must be a Dict[str, float], but got: {r}")

    for k, v in r.items():
        if not isinstance(k, str):
            raise TypeError(f"'{field_name}' keys must be strings, but got: {k}")
        if isinstance(v, (int, float)):
            if v < 0:
                raise ValueError(
                    f"'{field_name}' values must be >= 0, but got: '{k}: {v}'"
                )
        else:
            raise TypeError(
                f"'{field_name}' values must be floats, but got: '{k}: {v}'"
            )


def _validate_label_dict(labels: Optional[LabelDict]):
    if labels is None:
        return

    # Convert any non-string keys/values to strings to ensure compatibility
    for k, v in labels.items():
        if not isinstance(k, str):
            raise TypeError(f"'labels' keys must be strings, but got: {k}")
        if not isinstance(v, str):
            raise TypeError(f"'labels' values must be strings, but got: {v}")


def _validate_advanced_instance_config_dict(c: Optional[AdvancedInstanceConfigDict]):
    if c is None:
        return

    if not isinstance(c, dict) or not all(isinstance(k, str) for k in c):
        raise TypeError("'advanced_instance_config' must be a Dict[str, Any]")


@dataclass(frozen=True)
class CloudDeployment(ModelBase):
    """Cloud deployment selectors for a node group; one or more selectors may be passed to target a specific deployment from all of a cloud's deployments."""

    __doc_py_example__ = """
from anyscale.compute_config.models import CloudDeployment

cloud_deployment = CloudDeployment(
    provider="aws",
    region="us-west-2",
    machine_pool="machine-pool-name",
    id="cldrsrc_1234567890",
)
"""

    __doc_yaml_example__ = """
cloud_deployment:
  provider: aws
  region: us-west-2
  machine_pool: machine-pool-name
  id: cldrsrc_1234567890
"""

    provider: Optional[str] = field(
        default=None,
        metadata={"docstring": "Cloud provider name, e.g., `aws` or `gcp`."},
    )

    def _validate_provider(self, provider: Optional[str]):
        if provider is not None and not isinstance(provider, str):
            raise TypeError("'provider' must be a string.")

    region: Optional[str] = field(
        default=None,
        metadata={"docstring": "Cloud provider region, e.g., `us-west-2`."},
    )

    def _validate_region(self, region: Optional[str]):
        if region is not None and not isinstance(region, str):
            raise TypeError("'region' must be a string.")

    machine_pool: Optional[str] = field(
        default=None, metadata={"docstring": "Machine pool name."}
    )

    def _validate_machine_pool(self, machine_pool: Optional[str]):
        if machine_pool is not None and not isinstance(machine_pool, str):
            raise TypeError("'machine_pool' must be a string.")

    id: Optional[str] = field(
        default=None, metadata={"docstring": "Cloud deployment ID from cloud setup."}
    )

    def _validate_id(self, id: Optional[str]):  # noqa: A002
        if id is not None and not isinstance(id, str):
            raise TypeError("'id' must be a string.")


@dataclass(frozen=True)
class _NodeConfig(ModelBase):
    instance_type: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Cloud provider instance type, e.g., `m5.2xlarge` on AWS or `n2-standard-8` on GCP. "
            "Defaults to 'custom' when required_resources is provided."
        },
    )

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "_NodeConfig":
        """Create node config from dict, checking for deprecated physical_resources field."""
        # Check for deprecated physical_resources field
        if "physical_resources" in d and d["physical_resources"] is not None:
            raise ValueError(
                "'physical_resources' has been renamed to 'required_resources'. "
                "Please update your configuration to use 'required_resources' instead."
            )
        # Remove physical_resources if it's None to avoid passing unknown field
        if "physical_resources" in d:
            d = d.copy()
            del d["physical_resources"]
        return cls(**d)

    def _validate_instance_type(self, instance_type: Optional[str]) -> str:
        # Default to "custom" when required_resources is provided
        if instance_type is None:
            if self.required_resources is not None:
                instance_type = "custom"
            else:
                raise ValueError(
                    "'instance_type' is required when 'required_resources' is not provided."
                )
        if not isinstance(instance_type, str):
            raise TypeError("'instance_type' must be a string.")
        return instance_type

    def to_dict(self, *, exclude_none: bool = True) -> Dict[str, Any]:
        """Convert to dictionary, excluding 'instance_type: custom' for free pod shapes."""
        result = super().to_dict(exclude_none=exclude_none)

        # Remove instance_type if it's "custom" and required_resources is present
        # This makes the output cleaner for free pod shapes
        if (
            result.get("instance_type") == "custom"
            and result.get("required_resources") is not None
        ):
            del result["instance_type"]

        return result

    resources: Optional[ResourceDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Logical resources that will be available on this node. Defaults to match the physical resources of the instance type."
        },
    )

    def _validate_resources(self, resources: Optional[ResourceDict]):
        _validate_resource_dict(resources, field_name="resources")

    required_resources: Optional[PhysicalResources] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Physical resources for custom instance types (free pod shapes). Explicitly defines CPU, memory, and GPU resources."
        },
    )

    def _validate_required_resources(
        self, required_resources: Optional[PhysicalResources]
    ) -> Optional[PhysicalResources]:
        if required_resources is None:
            return None
        if isinstance(required_resources, dict):
            # Convert dict to PhysicalResources object
            return PhysicalResources.from_dict(required_resources)
        if isinstance(required_resources, PhysicalResources):
            return required_resources
        raise TypeError(
            "'required_resources' must be a PhysicalResources object or dict."
        )

    labels: Optional[LabelDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Labels to associate the node with for scheduling purposes. Defaults to the list of Ray & Anyscale default labels."
        },
    )

    def _validate_labels(self, labels: Optional[LabelDict]):
        _validate_label_dict(labels)

    required_labels: Optional[LabelDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Required labels that must be present on the node for scheduling purposes."
        },
    )

    def _validate_required_labels(self, required_labels: Optional[LabelDict]):
        _validate_label_dict(required_labels)
        # Cross-field validation: if accelerator-type is a GPU, GPU must be specified
        self._validate_gpu_accelerator_consistency(required_labels)
        # Cross-field validation: if accelerator-type is a TPU, TPU fields must be specified
        self._validate_tpu_accelerator_consistency(required_labels)

    def _validate_gpu_accelerator_consistency(
        self, required_labels: Optional[LabelDict]
    ):
        """Validate that GPU count is specified when a non-TPU accelerator type is requested.

        If ray.io/accelerator-type is specified and it's not a TPU type,
        then GPU must be specified in required_resources.
        """
        if required_labels is None:
            return

        accelerator_type = required_labels.get(RAY_TPU_ACCELERATOR_LABEL)
        if not accelerator_type:
            return

        # Skip if it's a TPU type (TPU validation is handled separately)
        if accelerator_type.upper().startswith("TPU"):
            return

        # Non-TPU accelerator type specified - assume it's a GPU and require GPU count
        # GPU accelerator type is specified, check if GPU count is provided
        if self.required_resources is None:
            raise ValueError(
                f"When specifying an accelerator type '{accelerator_type}' in "
                f"'required_labels', you must also specify 'GPU' count in 'required_resources'.\n"
                f"Example:\n"
                f"  required_resources:\n"
                f"    CPU: 7\n"
                f"    memory: 12Gi\n"
                f"    GPU: 1  # Required when using accelerator type\n"
                f"  required_labels:\n"
                f"    ray.io/accelerator-type: {accelerator_type}"
            )

        pr = self.required_resources
        gpu_count = (
            pr.get("GPU") or pr.get("gpu") or 0 if isinstance(pr, dict) else pr.GPU or 0
        )

        if not gpu_count or gpu_count <= 0:
            raise ValueError(
                f"When specifying an accelerator type '{accelerator_type}' in "
                f"'required_labels', you must also specify 'GPU' count in 'required_resources'.\n"
                f"Example:\n"
                f"  required_resources:\n"
                f"    CPU: 7\n"
                f"    memory: 12Gi\n"
                f"    GPU: 1  # Required when using accelerator type\n"
                f"  required_labels:\n"
                f"    ray.io/accelerator-type: {accelerator_type}"
            )

    def _validate_tpu_accelerator_consistency(
        self, required_labels: Optional[LabelDict]
    ):
        """Validate that TPU fields are specified when a TPU accelerator type is requested.

        If ray.io/accelerator-type is a TPU type (starts with "TPU"),
        then TPU, tpu_hosts, and ray.io/tpu-topology must be specified.
        """
        if required_labels is None:
            return

        accelerator_type = required_labels.get(RAY_TPU_ACCELERATOR_LABEL)
        if not accelerator_type:
            return

        # Only validate if it's a TPU type
        if not accelerator_type.upper().startswith("TPU"):
            return

        # TPU accelerator type specified - check required fields
        missing_fields = []

        # Check TPU count in required_resources
        tpu_count = 0
        tpu_hosts = 0
        if self.required_resources is not None:
            pr = self.required_resources
            if isinstance(pr, dict):
                tpu_count = pr.get("TPU") or pr.get("tpu") or 0
                tpu_hosts = pr.get("tpu_hosts") or 0
            else:
                tpu_count = pr.TPU or 0
                tpu_hosts = pr.tpu_hosts or 0

        if not tpu_count or tpu_count <= 0:
            missing_fields.append("TPU (in required_resources)")

        if not tpu_hosts or tpu_hosts <= 0:
            missing_fields.append("tpu_hosts (in required_resources)")

        # Check ray.io/tpu-topology in required_labels or labels
        has_topology = RAY_TPU_TOPOLOGY_LABEL in required_labels
        if not has_topology and self.labels:
            has_topology = RAY_TPU_TOPOLOGY_LABEL in self.labels

        if not has_topology:
            missing_fields.append("ray.io/tpu-topology (in required_labels)")

        if missing_fields:
            raise ValueError(
                f"When specifying a TPU accelerator type '{accelerator_type}' in "
                f"'required_labels', you must also specify: {', '.join(missing_fields)}.\n"
                f"Example:\n"
                f"  required_resources:\n"
                f"    CPU: 7\n"
                f"    memory: 12Gi\n"
                f"    TPU: 4\n"
                f"    tpu_hosts: 1\n"
                f"  required_labels:\n"
                f"    ray.io/accelerator-type: {accelerator_type}\n"
                f"    ray.io/tpu-topology: 2x2\n"
                f"For more information, see: {TPU_DOCS_URL}"
            )

    advanced_instance_config: Optional[AdvancedInstanceConfigDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Advanced instance configurations that will be passed through to the cloud provider.",
            "customer_hosted_only": True,
        },
    )

    def _validate_advanced_instance_config(
        self, advanced_instance_config: Optional[AdvancedInstanceConfigDict]
    ):
        _validate_advanced_instance_config_dict(advanced_instance_config)
        # Validate TPU-specific node selectors if TPU is configured
        self._validate_tpu_node_selectors(advanced_instance_config)
        # Validate accelerator type consistency between required_labels and nodeSelector
        self._validate_accelerator_consistency(advanced_instance_config)

    def _validate_accelerator_consistency(
        self, advanced_instance_config: Optional[AdvancedInstanceConfigDict]
    ):
        """Validate that accelerator types are consistent between required_labels and nodeSelector.

        If both ray.io/accelerator-type (in required_labels) and cloud.google.com/gke-accelerator
        (in nodeSelector) are specified, they should refer to the same accelerator type.
        """
        if advanced_instance_config is None:
            return

        # Extract nodeSelector from advanced_instance_config
        spec = advanced_instance_config.get("spec", {})
        if not isinstance(spec, dict):
            return
        node_selectors = spec.get("nodeSelector", {})
        if not node_selectors:
            return

        # Get cloud.google.com/gke-accelerator from nodeSelector
        gke_accelerator = node_selectors.get("cloud.google.com/gke-accelerator")
        if not gke_accelerator:
            return

        # Get ray.io/accelerator-type from required_labels or labels
        ray_accelerator = None
        if self.required_labels and RAY_TPU_ACCELERATOR_LABEL in self.required_labels:
            ray_accelerator = self.required_labels[RAY_TPU_ACCELERATOR_LABEL]
        elif self.labels and RAY_TPU_ACCELERATOR_LABEL in self.labels:
            ray_accelerator = self.labels[RAY_TPU_ACCELERATOR_LABEL]

        if not ray_accelerator:
            return

        # Both are specified - check for consistency
        # Normalize gke-accelerator by removing common prefixes and extracting GPU identifier
        # e.g., "nvidia-tesla-t4" -> "t4", "nvidia-a100-80gb" -> "a100"
        gke_normalized = gke_accelerator.lower()
        for prefix in ["nvidia-", "tesla-", "geforce-", "quadro-"]:
            gke_normalized = gke_normalized.replace(prefix, "")
        # Remove common suffixes like "-80gb", "-40gb", "-sxm", "-pcie"
        for suffix in ["-80gb", "-40gb", "-sxm4", "-sxm", "-pcie", "-nvlink"]:
            gke_normalized = gke_normalized.replace(suffix, "")
        gke_normalized = gke_normalized.strip("-").replace("-", "").replace("_", "")

        # Normalize ray accelerator type
        ray_normalized = ray_accelerator.lower().replace("-", "").replace("_", "")

        # Check if there's any overlap - the normalized values should match or one should contain the other
        # This handles cases like:
        #   - ray: "T4" vs gke: "nvidia-tesla-t4" -> both normalize to contain "t4"
        #   - ray: "A100" vs gke: "nvidia-a100-80gb" -> both normalize to contain "a100"
        is_consistent = (
            gke_normalized in ray_normalized
            or ray_normalized in gke_normalized
            or gke_normalized == ray_normalized
        )

        if not is_consistent:
            warnings.warn(
                f"Potentially inconsistent accelerator types: 'ray.io/accelerator-type: {ray_accelerator}' "
                f"may not match 'cloud.google.com/gke-accelerator: {gke_accelerator}' in nodeSelector. "
                f"Please ensure both refer to the same GPU type.",
                UserWarning,
                stacklevel=2,
            )

    def _validate_tpu_node_selectors(
        self, advanced_instance_config: Optional[AdvancedInstanceConfigDict]
    ):
        """Validate that required GKE TPU configuration is present when TPU is specified.

        When using TPUs on GKE, the following are required:
        - tpu_hosts: Specifies single-host (1) or multi-host (>1) TPU configuration
        - Either explicit node selectors OR labels/required_labels that can be used to derive them:
          Option 1 (explicit node selectors):
            - cloud.google.com/gke-tpu-topology: Specifies the TPU topology (e.g., '2x2', '4x4')
            - cloud.google.com/gke-tpu-accelerator: Specifies the TPU type (e.g., 'tpu-v5-lite-podslice')
          Option 2 (labels or required_labels - node selectors will be derived):
            - ray.io/accelerator-type: TPU accelerator type (e.g., 'TPU-V6E')
            - ray.io/tpu-topology: TPU topology (e.g., '2x2')

        Note: TPU labels must not be specified in both labels and required_labels.

        See: https://cloud.google.com/kubernetes-engine/docs/concepts/plan-tpus
        """
        # Check if TPU is configured in required_resources
        if self.required_resources is None:
            return

        pr = self.required_resources
        if isinstance(pr, dict):
            tpu_count = pr.get("TPU") or pr.get("tpu") or 0
            tpu_hosts = pr.get("tpu_hosts") or 0
        else:
            tpu_count = pr.TPU or 0
            tpu_hosts = pr.tpu_hosts or 0

        if not tpu_count or tpu_count <= 0:
            return

        # TPU is configured, validate tpu_hosts is specified
        if not tpu_hosts or tpu_hosts <= 0:
            raise ValueError(
                f"TPU configuration requires 'tpu_hosts' in 'required_resources' to specify "
                f"single-host (tpu_hosts: 1) or multi-host (tpu_hosts: >1) TPU configuration. "
                f"Example:\n"
                f"  required_resources:\n"
                f"    CPU: 7\n"
                f"    memory: 12Gi\n"
                f"    TPU: 4\n"
                f"    accelerator: TPU-V5E\n"
                f"    tpu_hosts: 4  # Required for TPU\n"
                f"For more information on TPU configuration, see: {TPU_DOCS_URL}"
            )

        # Extract node selectors from advanced_instance_config
        node_selectors = {}
        if advanced_instance_config:
            spec = advanced_instance_config.get("spec", {})
            if isinstance(spec, dict):
                node_selectors = spec.get("nodeSelector", {})

        # Check if labels can be used to derive the node selectors
        labels = self.labels or {}
        required_labels = self.required_labels or {}

        # Helper to check for TPU labels in a label dict
        def has_tpu_labels(label_dict: Dict[str, str]) -> tuple:
            has_accel = RAY_TPU_ACCELERATOR_LABEL in label_dict and label_dict[
                RAY_TPU_ACCELERATOR_LABEL
            ].upper().startswith("TPU")
            has_topo = RAY_TPU_TOPOLOGY_LABEL in label_dict
            return has_accel, has_topo

        labels_accel, labels_topo = has_tpu_labels(labels)
        req_labels_accel, req_labels_topo = has_tpu_labels(required_labels)

        # Check for conflicts: TPU labels should not be in both labels and required_labels
        conflicting_labels = []
        if labels_accel and req_labels_accel:
            conflicting_labels.append(RAY_TPU_ACCELERATOR_LABEL)
        if labels_topo and req_labels_topo:
            conflicting_labels.append(RAY_TPU_TOPOLOGY_LABEL)

        if conflicting_labels:
            conflict_str = ", ".join(f"'{label}'" for label in conflicting_labels)
            raise ValueError(
                f"TPU configuration labels {conflict_str} cannot be specified in both "
                f"'labels' and 'required_labels'. Please specify them in only one place."
            )

        # If both TPU labels are present (in either labels or required_labels),
        # node selectors will be derived by the backend
        has_tpu_accelerator_label = labels_accel or req_labels_accel
        has_tpu_topology_label = labels_topo or req_labels_topo

        if has_tpu_accelerator_label and has_tpu_topology_label:
            return

        # Check for missing node selectors (only if labels are not provided)
        missing_selectors = [
            selector
            for selector in REQUIRED_TPU_NODE_SELECTORS
            if selector not in node_selectors
        ]

        if missing_selectors:
            missing_str = ", ".join(f"'{s}'" for s in missing_selectors)
            raise ValueError(
                f"TPU configuration requires either:\n"
                f"1. Node selectors in 'advanced_instance_config.spec.nodeSelector': {missing_str}\n"
                f"   Example:\n"
                f"     advanced_instance_config:\n"
                f"       spec:\n"
                f"         nodeSelector:\n"
                f"           cloud.google.com/gke-tpu-topology: '4x4'\n"
                f"           cloud.google.com/gke-tpu-accelerator: 'tpu-v5-lite-podslice'\n"
                f"2. OR labels/required_labels that will be used to derive node selectors:\n"
                f"   Example:\n"
                f"     labels:\n"
                f"       ray.io/accelerator-type: TPU-V6E\n"
                f"       ray.io/tpu-topology: 2x2\n"
                f"For more information on TPU configuration, see: {TPU_DOCS_URL}"
            )

    flags: Optional[Dict[str, Any]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Node-level flags specifying advanced or experimental options.",
            "customer_hosted_only": False,
        },
    )

    def _validate_flags(self, flags: Optional[Dict[str, Any]]):
        if flags is None:
            return

        if not isinstance(flags, dict):
            raise TypeError("'flags' must be a dict")

    cloud_deployment: Union[CloudDeployment, Dict[str, str], None] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Cloud deployment selectors for a node group; one or more selectors may be passed to target a specific deployment from all of a cloud's deployments.",
            "customer_hosted_only": False,
        },
    )

    def _validate_cloud_deployment(
        self, cloud_deployment: Union[CloudDeployment, Dict[str, str], None]
    ) -> Optional[CloudDeployment]:
        if cloud_deployment is None:
            return None
        if isinstance(cloud_deployment, dict):
            cloud_deployment = CloudDeployment.from_dict(cloud_deployment)
        if not isinstance(cloud_deployment, CloudDeployment):
            raise TypeError(
                "'cloud_deployment' must be a CloudDeployment or corresponding dict"
            )
        return cloud_deployment


@dataclass(frozen=True)
class HeadNodeConfig(_NodeConfig):
    """Configuration options for the head node of a cluster."""

    __doc_py_example__ = """
from anyscale.compute_config.models import ComputeConfig, HeadNodeConfig

config = ComputeConfig(
    head_node=HeadNodeConfig(
        instance_type="m5.8xlarge",
    ),
)
"""

    __doc_yaml_example__ = """
head_node:
  instance_type: m5.8xlarge
"""


class MarketType(ModelEnum):
    """Market type of instances to use (on-demand vs. spot)."""

    ON_DEMAND = "ON_DEMAND"
    SPOT = "SPOT"
    PREFER_SPOT = "PREFER_SPOT"

    __docstrings__: ClassVar[Dict[str, str]] = {
        ON_DEMAND: "Use on-demand instances only.",
        SPOT: "Use spot instances only.",
        PREFER_SPOT: (
            "Prefer to use spot instances, but fall back to on-demand if necessary. "
            "If on-demand instances are running and spot instances become available, "
            "the on-demand instances will be evicted and replaced with spot instances."
        ),
    }


@dataclass(frozen=True)
class WorkerNodeGroupConfig(_NodeConfig):
    """Configuration options for a worker node group in a cluster.

    Clusters can have multiple worker node groups that use different instance types or configurations.
    """

    __doc_py_example__ = """
from anyscale.compute_config.models import ComputeConfig, MarketType, WorkerNodeGroupConfig

config = ComputeConfig(
    worker_nodes=[
        WorkerNodeGroupConfig(
            instance_type="m5.8xlarge",
            min_nodes=5,
            max_nodes=5,
        ),
        WorkerNodeGroupConfig(
            instance_type="m5.4xlarge",
            min_nodes=1,
            max_nodes=10,
            market_type=MarketType.SPOT,
        ),
    ],
)
"""

    __doc_yaml_example__ = """
worker_nodes:
- instance_type: m5.8xlarge
  min_nodes: 5
  max_nodes: 5
- instance_type: m5.4xlarge
  min_nodes: 1
  max_nodes: 10
  market_type: SPOT
"""

    name: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Unique name of this worker group. Defaults to a human-friendly representation of the instance type."
        },
    )

    def _validate_name(self, name: Optional[str]) -> str:
        # Default name to the instance type if not specified.
        if name is None:
            # For free pod shapes (custom instance types), generate a unique name
            # from required resources to allow multiple custom worker groups
            if self.instance_type == "custom" and self.required_resources is not None:
                pr = self.required_resources
                # Get accelerator from required_resources, or fall back to
                # required_labels/labels if not specified (ray.io/accelerator-type)
                accelerator = pr.accelerator
                if not accelerator:
                    accelerator = _extract_accelerator_from_labels(
                        self.labels, self.required_labels
                    )
                name = _generate_free_pod_shape_name(
                    cpu=pr.CPU,
                    memory=pr.memory,
                    gpu=pr.GPU,
                    accelerator=accelerator,
                    tpu=pr.TPU,
                    tpu_hosts=pr.tpu_hosts,
                    cpu_architecture=pr.cpu_architecture,
                )
            else:
                name = self.instance_type

        if not isinstance(name, str):
            raise TypeError("'name' must be a string")
        if len(name) == 0:
            raise ValueError("'name' cannot be empty")

        return name

    min_nodes: int = field(
        default=0,
        metadata={
            "docstring": "Minimum number of nodes of this type that will be kept running in the cluster."
        },
    )

    def _validate_min_nodes(self, min_nodes: int):
        if not isinstance(min_nodes, int):
            raise TypeError("'min_nodes' must be an int")
        if min_nodes < 0:
            raise ValueError("'min_nodes' must be >= 0")

    max_nodes: int = field(
        default=10,
        metadata={
            "docstring": "Maximum number of nodes of this type that can be running in the cluster."
        },
    )

    def _validate_max_nodes(self, max_nodes: int):
        if not isinstance(max_nodes, int):
            raise TypeError("'max_nodes' must be an int")
        if max_nodes < 1:
            raise ValueError("'max_nodes' must be >= 1")
        if max_nodes < self.min_nodes:
            raise ValueError(f"'max_nodes' must be >= 'min_nodes' ({self.min_nodes})")

    market_type: Union[str, MarketType] = field(
        default=MarketType.ON_DEMAND,
        metadata={
            "docstring": "The type of instances to use (see `MarketType` enum values for details).",
            "customer_hosted_only": True,
        },
    )

    def _validate_market_type(self, market_type: Union[str, MarketType]) -> MarketType:
        if isinstance(market_type, str):
            # This will raise a ValueError if the market_type is unrecognized.
            market_type = MarketType(market_type)
        elif not isinstance(market_type, MarketType):
            raise TypeError("'market_type' must be a MarketType.")

        return market_type


@dataclass(frozen=True)
class ComputeConfig(ModelBase):
    """Compute configuration for instance types and cloud resources for a cluster with a single cloud resource."""

    __doc_py_example__ = """
from anyscale.compute_config.models import (
    ComputeConfig, HeadNodeConfig, MarketType, WorkerNodeGroupConfig
)

config = ComputeConfig(
    cloud="my-cloud",
    head_node=HeadNodeConfig(
        instance_type="m5.8xlarge",
    ),
    worker_nodes=[
        WorkerNodeGroupConfig(
            instance_type="m5.8xlarge",
            min_nodes=5,
            max_nodes=5,
        ),
        WorkerNodeGroupConfig(
            instance_type="m5.4xlarge",
            min_nodes=1,
            max_nodes=10,
            market_type=MarketType.SPOT,
        ),
    ],
)
"""

    __doc_yaml_example__ = """
cloud: my-cloud
zones: # (Optional) Defaults to to all zones in a region.
  - us-west-2a
  - us-west-2b
head_node:
  instance_type: m5.8xlarge
worker_nodes:
- instance_type: m5.8xlarge
  min_nodes: 5
  max_nodes: 5
  market_type: PREFER_SPOT # (Optional) Defaults to ON_DEMAND
- instance_type: g5.4xlarge
  min_nodes: 1
  max_nodes: 10
  market_type: SPOT # (Optional) Defaults to ON_DEMAND
min_resources: # (Optional) Defaults to no minimum.
  CPU: 1
  GPU: 1
  CUSTOM_RESOURCE: 0
max_resources: # (Optional) Defaults to no maximum.
  CPU: 6
  GPU: 10
  CUSTOM_RESOURCE: 10
enable_cross_zone_scaling: true # (Optional) Defaults to false.
advanced_instance_config: # (Optional) Defaults to no advanced configurations.
  # AWS specific configuration example
  BlockDeviceMappings:
    - DeviceName: DEVICE_NAME
      Ebs:
        VolumeSize: VOLUME_SIZE
        DeleteOnTermination: DELETE_ON_TERMINATION
  IamInstanceProfile:
    Arn: IAM_INSTANCE_PROFILE_ARN
  NetworkInterfaces:
    - SubnetId: SUBNET_ID
      Groups:
        - SECURITY_GROUP_ID
      AssociatePublicIpAddress: ASSOCIATE_PUBLIC_IP
  TagSpecifications:
    - ResourceType: RESOURCE_TYPE
      Tags:
        - Key: TAG_KEY
          Value: TAG_VALUE
  # GCP specific configuration example
  instance_properties:
    disks:
      - boot: BOOT_OPTION
        auto_delete: AUTO_DELETE_OPTION
        initialize_params:
          disk_size_gb: DISK_SIZE_GB
    service_accounts:
      - email: SERVICE_ACCOUNT_EMAIL
        scopes:
          - SCOPE_URL
    network_interfaces:
      - subnetwork: SUBNETWORK_URL
        access_configs:
          - type: ACCESS_CONFIG_TYPE
    labels:
      LABEL_KEY: LABEL_VALUE
"""

    cloud: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace)."
        },
    )

    def _validate_cloud(self, cloud: Optional[str]):
        if cloud is not None and not isinstance(cloud, str):
            raise TypeError("'cloud' must be a string")

    cloud_resource: Optional[str] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "The cloud resource to use for this workload. Defaults to the primary cloud resource of the Cloud.",
            "customer_hosted_only": True,
        },
    )

    def _validate_cloud_resource(self, cloud_resource: Optional[str]):
        if cloud_resource is not None and not isinstance(cloud_resource, str):
            raise TypeError("'cloud_resource' must be a string")

    head_node: Union[HeadNodeConfig, Dict, None] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Configuration options for the head node of the cluster. Defaults to the cloud's default head node configuration."
        },
    )

    def _validate_head_node(
        self, head_node: Union[HeadNodeConfig, Dict, None]
    ) -> Optional[HeadNodeConfig]:
        if head_node is None:
            return None

        if isinstance(head_node, dict):
            head_node = HeadNodeConfig.from_dict(head_node)
        if not isinstance(head_node, HeadNodeConfig):
            raise TypeError(
                "'head_node' must be a HeadNodeConfig or corresponding dict"
            )

        return head_node

    worker_nodes: Optional[List[Union[WorkerNodeGroupConfig, Dict]]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Configuration options for the worker nodes of the cluster. If not provided, worker nodes will be automatically selected based on logical resource requests. To use a head-node only cluster, pass `[]` here."
        },
    )

    def _validate_worker_nodes(
        self, worker_nodes: Optional[List[Union[WorkerNodeGroupConfig, Dict]]]
    ) -> Optional[List[WorkerNodeGroupConfig]]:
        if worker_nodes is None:
            return None

        if not isinstance(worker_nodes, list) or not all(
            isinstance(c, (dict, WorkerNodeGroupConfig)) for c in worker_nodes
        ):
            raise TypeError(
                "'worker_nodes' must be a list of WorkerNodeGroupConfigs or corresponding dicts"
            )

        duplicate_names = set()
        name_counts: DefaultDict[str, int] = defaultdict(int)
        worker_node_models: List[WorkerNodeGroupConfig] = []
        for node in worker_nodes:
            parsed_node = (
                WorkerNodeGroupConfig.from_dict(node)
                if isinstance(node, dict)
                else node
            )
            assert isinstance(parsed_node, WorkerNodeGroupConfig)
            worker_node_models.append(parsed_node)
            name = parsed_node.name
            assert name is not None
            name_counts[name] += 1
            if name_counts[name] > 1:
                duplicate_names.add(name)

        if duplicate_names:
            raise ValueError(
                f"'worker_nodes' names must be unique, but got duplicate names: {duplicate_names}"
            )

        return worker_node_models

    min_resources: Optional[ResourceDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Total minimum logical resources across all nodes in the cluster. Resources omitted from this field have no minimum.",
            "customer_hosted_only": False,
        },
    )

    def _validate_min_resources(self, min_resources: Optional[ResourceDict]):
        _validate_resource_dict(min_resources, field_name="min_resources")

    max_resources: Optional[ResourceDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Total maximum logical resources across all nodes in the cluster. Resources omitted from this field have no maximum.",
            "customer_hosted_only": False,
        },
    )

    def _validate_max_resources(self, max_resources: Optional[ResourceDict]):
        _validate_resource_dict(max_resources, field_name="max_resources")

    zones: Optional[List[str]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Availability zones to consider for this cluster. Defaults to all zones in the cloud's region. By default all instances with user workloads scheduled on them will run in the same zone to save cost, unless `enable_cross_zone_scaling` is set.",
            "customer_hosted_only": True,
        },
    )

    def _validate_zones(self, zones: Optional[List[str]]):
        if zones is None:
            return
        if not isinstance(zones, list) or not all(isinstance(z, str) for z in zones):
            raise TypeError("'zones' must be a List[str]")
        if len(zones) == 0:
            raise ValueError(
                "'zones' must not be an empty list. Set `None` to default to all zones."
            )

    enable_cross_zone_scaling: bool = field(
        default=False,
        repr=False,
        metadata={
            "docstring": "Allow instances in the cluster to be run across multiple zones. This is recommended when running production services (for fault-tolerance in a zone failure scenario). It is not recommended for workloads that have a large amount of inter-zone communication due to the possibility of higher costs and degraded performance. When false, all instances with user workloads scheduled on them (e.g. all worker nodes in multi-node clusters) will run in the same zone to save cost.",
            "customer_hosted_only": True,
        },
    )

    def _validate_enable_cross_zone_scaling(self, enable_cross_zone_scaling: bool):
        if not isinstance(enable_cross_zone_scaling, bool):
            raise TypeError("'enable_cross_zone_scaling' must be a boolean")

    advanced_instance_config: Optional[AdvancedInstanceConfigDict] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Advanced instance configurations that will be passed through to the cloud provider.",
            "customer_hosted_only": True,
        },
    )

    def _validate_advanced_instance_config(
        self, advanced_instance_config: Optional[AdvancedInstanceConfigDict],
    ):
        _validate_advanced_instance_config_dict(advanced_instance_config)

    flags: Optional[Dict[str, Any]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Cluster-level flags specifying advanced or experimental options.",
            "customer_hosted_only": False,
        },
    )

    def _validate_flags(self, flags: Optional[Dict[str, Any]]):
        if flags is None:
            return

        if not isinstance(flags, dict):
            raise TypeError("'flags' must be a dict")

    auto_select_worker_config: bool = field(
        default=False,
        repr=False,
        metadata={
            "docstring": "Allow worker groups to be automatically configured based on the workload's logical resource requests. When false, worker groups must be explicitly configured.",
        },
    )

    def _validate_auto_select_worker_config(self, auto_select_worker_config: bool):
        if not isinstance(auto_select_worker_config, bool):
            raise TypeError("'auto_select_worker_config' must be a boolean")
        if auto_select_worker_config and self.worker_nodes is not None:
            raise ValueError(
                "'auto_select_worker_config' must be false when 'worker_nodes' are provided"
            )


@dataclass(frozen=True)
class MultiResourceComputeConfig(ModelBase):
    """EXPERIMENTAL. Compute configuration for a cluster with multiple possible cloud resources."""

    __doc_py_example__ = """
from anyscale.compute_config.models import (
    MultiResourceComputeConfig, ComputeConfig, HeadNodeConfig, WorkerNodeGroupConfig
)
config = MultiResourceComputeConfig(
    cloud="my-cloud",
    configs=[
        ComputeConfig(
            cloud_resource="vm-aws-us-west-1",
            head_node=HeadNodeConfig(
                instance_type="m5.2xlarge",
            ),
            worker_nodes=[
                WorkerNodeGroupConfig(
                    instance_type="m5.4xlarge",
                    min_nodes=1,
                    max_nodes=10,
                ),
            ],
        ),
        ComputeConfig(
            cloud_resource="vm-aws-us-west-2",
            head_node=HeadNodeConfig(
                instance_type="m5.2xlarge",
            ),
            worker_nodes=[
                WorkerNodeGroupConfig(
                    instance_type="m5.4xlarge",
                    min_nodes=1,
                    max_nodes=10,
                ),
            ],
        )
    ]
)
"""

    __doc_yaml_example__ = """
cloud: my-cloud
configs:
- cloud_resource: vm-aws-us-west-1
  head_node:
    instance_type: m5.2xlarge
  worker_nodes:
  - instance_type: m5.4xlarge
    min_nodes: 1
    max_nodes: 10
- cloud_resource: vm-aws-us-west-2
  head_node:
    instance_type: m5.2xlarge
  worker_nodes:
  - instance_type: m5.4xlarge
    min_nodes: 1
    max_nodes: 10
"""
    cloud: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace)."
        },
    )

    def _validate_cloud(self, cloud: Optional[str]):
        if cloud is not None and not isinstance(cloud, str):
            raise TypeError("'cloud' must be a string")

    configs: List[Union[ComputeConfig, Dict]] = field(
        default_factory=list,
        repr=False,
        metadata={
            "docstring": "List of compute configurations, one for each cloud resource.",
            "customer_hosted_only": True,
        },
    )

    def _validate_configs(
        self, configs: List[Union[ComputeConfig, Dict]]
    ) -> List[ComputeConfig]:
        if not isinstance(configs, list) or not all(
            isinstance(c, (dict, ComputeConfig)) for c in configs
        ):
            raise TypeError(
                "'configs' must be a list of ComputeConfigs or corresponding dicts"
            )

        config_models: List[ComputeConfig] = []
        unique_clouds = set()
        unique_resources = set()
        for config in configs:
            parsed_config = (
                ComputeConfig.from_dict(config) if isinstance(config, dict) else config
            )
            assert isinstance(parsed_config, ComputeConfig)
            config_models.append(parsed_config)

            if parsed_config.cloud:
                unique_clouds.add(parsed_config.cloud)

            unique_resources.add(parsed_config.cloud_resource)

        if len(unique_clouds) > 1:
            raise ValueError("'cloud' must be the same for all configs.")

        if len(unique_resources) != len(configs):
            raise ValueError(
                "'cloud_resource' must be unique for each compute configuration."
            )

        if len(configs) == 0:
            raise ValueError(
                "'configs' must include at least one compute configuration."
            )

        return config_models

    flags: Optional[Dict[str, Any]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Flags specifying advanced or experimental options that should be applied to all cloud resources. Flags specified in the individual compute configurations will override these flags.",
        },
    )

    def _validate_flags(self, flags: Optional[Dict[str, Any]]):
        if flags is None:
            return

        if not isinstance(flags, dict):
            raise TypeError("'flags' must be a dict")


ComputeConfigType = Union[ComputeConfig, MultiResourceComputeConfig]


def compute_config_type_from_yaml(config_file: str) -> ComputeConfigType:
    """
    Parse a YAML compute config file into either a ComputeConfig or MultiResourceComputeConfig.
    """
    error_message = f"Could not parse config file '{config_file}' as a ComputeConfig or MultiResourceComputeConfig:\n"

    try:
        return ComputeConfig.from_yaml(config_file)
    except Exception as e:  # noqa: BLE001
        error_message += f"ComputeConfig: {e}\n"

    try:
        return MultiResourceComputeConfig.from_yaml(config_file)
    except Exception as e:  # noqa: BLE001
        error_message += f"MultiResourceComputeConfig: {e}\n"

    raise TypeError(error_message.rstrip())


def compute_config_type_from_dict(config_dict: Dict) -> ComputeConfigType:
    """
    Parse a compute config dict into either a ComputeConfig or MultiResourceComputeConfig.
    """
    error_message = f"Could not parse config dict '{config_dict}' as a ComputeConfig or MultiResourceComputeConfig:\n"

    try:
        return ComputeConfig.from_dict(config_dict)
    except Exception as e:  # noqa: BLE001
        error_message += f"ComputeConfig: {e}\n"

    try:
        return MultiResourceComputeConfig.from_dict(config_dict)
    except Exception as e:  # noqa: BLE001
        error_message += f"MultiResourceComputeConfig: {e}\n"

    raise TypeError(error_message.rstrip())


@dataclass(frozen=True)
class ComputeConfigVersion(ModelBase):
    """Details of a created version of a compute config.

    Includes the config options and metadata such as the name, version, and ID.
    """

    __doc_py_example__ = """
import anyscale
from anyscale.compute_config.models import (
    ComputeConfigVersion
)

version: ComputeConfigVersion = anyscale.compute_config.get("my-compute-config")
"""

    __doc_cli_example__ = """\
$ anyscale compute-config get -n my-compute-config
name: my-compute-config:1
id: cpt_r4b4b3621rl3uggg7llj3mvme6
config:
  cloud: my-cloud
  head_node:
    instance_type: m5.8xlarge
  worker_nodes:
  - instance_type: m5.8xlarge
    min_nodes: 5
    max_nodes: 5
  - instance_type: m5.4xlarge
    min_nodes: 1
    max_nodes: 10
    market_type: SPOT
"""

    name: str = field(
        metadata={
            "docstring": "Name of the compute config including the version tag, i.e., 'name:version'."
        }
    )

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("'name' must be a string.")

        if not name.count(":") == 1:
            raise ValueError("'name' must be in the format: '<name>:<version>'.")

    id: str = field(metadata={"docstring": "Unique ID of the compute config."})

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("'id' must be a string.")

    config: Optional[ComputeConfigType] = field(
        default=None, metadata={"docstring": "The compute configuration."},
    )

    def _validate_config(self, config: Optional[ComputeConfigType]):
        if config is not None and not isinstance(
            config, (ComputeConfig, MultiResourceComputeConfig)
        ):
            raise TypeError(
                "'config' must be a ComputeConfig or MultiResourceComputeConfig"
            )


@dataclass(frozen=True)
class ComputeConfigListResult(ModelBase):
    """Result object returned by anyscale.compute_config.list().

    Contains a list of compute configs matching the query, along with pagination metadata.
    """

    __doc_py_example__ = """
import anyscale
from anyscale.compute_config.models import ComputeConfigListResult

# List compute configs with pagination
result: ComputeConfigListResult = anyscale.compute_config.list(
    cloud_name="aws-prod",
    sort_by="created_at",
    max_items=10
)

# Access results
print(f"Found {result.count} compute configs")
for config in result.results:
    print(f"  - {config.name}:{config.version}")

# Handle pagination
if result.next_token:
    next_page = anyscale.compute_config.list(
        cloud_name="aws-prod",
        max_items=10,
        next_token=result.next_token
    )
"""

    results: List[Any] = field(
        default_factory=list,
        metadata={
            "docstring": "List of compute config objects matching the query. Each item is a ClusterCompute object with details about the compute configuration."
        },
    )

    def _validate_results(self, results: List[Any]):
        if not isinstance(results, list):
            raise TypeError("'results' must be a list.")

    next_token: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Pagination token for fetching the next page of results. If None, there are no more results available."
        },
    )

    def _validate_next_token(self, next_token: Optional[str]):
        if next_token is not None and not isinstance(next_token, str):
            raise TypeError("'next_token' must be a string or None.")

    count: int = field(
        default=0,
        metadata={
            "docstring": "Number of results returned in this response. Useful for checking if there are results without iterating through the list."
        },
    )

    def _validate_count(self, count: int):
        if not isinstance(count, int):
            raise TypeError("'count' must be an integer.")
        if count < 0:
            raise ValueError("'count' must be non-negative.")



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ClusterComputesQuery(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'project_id': 'str',
        'creator_id': 'str',
        'name': 'TextQuery',
        'include_anonymous': 'bool',
        'paging': 'PageQuery',
        'cloud_id': 'str',
        'version': 'int',
        'sort_by_clauses': 'list[SortByClauseClusterComputesSortField]'
    }

    attribute_map = {
        'project_id': 'project_id',
        'creator_id': 'creator_id',
        'name': 'name',
        'include_anonymous': 'include_anonymous',
        'paging': 'paging',
        'cloud_id': 'cloud_id',
        'version': 'version',
        'sort_by_clauses': 'sort_by_clauses'
    }

    def __init__(self, project_id=None, creator_id=None, name=None, include_anonymous=False, paging=None, cloud_id=None, version=None, sort_by_clauses=[{"sort_field":"CREATED_AT","sort_order":"DESC"},{"sort_field":"NAME","sort_order":"ASC"},{"sort_field":"ID","sort_order":"ASC"}], local_vars_configuration=None):  # noqa: E501
        """ClusterComputesQuery - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._project_id = None
        self._creator_id = None
        self._name = None
        self._include_anonymous = None
        self._paging = None
        self._cloud_id = None
        self._version = None
        self._sort_by_clauses = None
        self.discriminator = None

        if project_id is not None:
            self.project_id = project_id
        if creator_id is not None:
            self.creator_id = creator_id
        if name is not None:
            self.name = name
        if include_anonymous is not None:
            self.include_anonymous = include_anonymous
        if paging is not None:
            self.paging = paging
        if cloud_id is not None:
            self.cloud_id = cloud_id
        if version is not None:
            self.version = version
        if sort_by_clauses is not None:
            self.sort_by_clauses = sort_by_clauses

    @property
    def project_id(self):
        """Gets the project_id of this ClusterComputesQuery.  # noqa: E501

        Filters Cluster Computes by project. If this field is absent, no filtering is done.  # noqa: E501

        :return: The project_id of this ClusterComputesQuery.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this ClusterComputesQuery.

        Filters Cluster Computes by project. If this field is absent, no filtering is done.  # noqa: E501

        :param project_id: The project_id of this ClusterComputesQuery.  # noqa: E501
        :type: str
        """

        self._project_id = project_id

    @property
    def creator_id(self):
        """Gets the creator_id of this ClusterComputesQuery.  # noqa: E501

        Filters Cluster Computes by creator. If this field is absent, no filtering is done.  # noqa: E501

        :return: The creator_id of this ClusterComputesQuery.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this ClusterComputesQuery.

        Filters Cluster Computes by creator. If this field is absent, no filtering is done.  # noqa: E501

        :param creator_id: The creator_id of this ClusterComputesQuery.  # noqa: E501
        :type: str
        """

        self._creator_id = creator_id

    @property
    def name(self):
        """Gets the name of this ClusterComputesQuery.  # noqa: E501

        Filters Cluster Computes by name. If this field is absent, no filtering is done.  # noqa: E501

        :return: The name of this ClusterComputesQuery.  # noqa: E501
        :rtype: TextQuery
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ClusterComputesQuery.

        Filters Cluster Computes by name. If this field is absent, no filtering is done.  # noqa: E501

        :param name: The name of this ClusterComputesQuery.  # noqa: E501
        :type: TextQuery
        """

        self._name = name

    @property
    def include_anonymous(self):
        """Gets the include_anonymous of this ClusterComputesQuery.  # noqa: E501

        Whether to include anonymous Cluster Computes in the search.  # noqa: E501

        :return: The include_anonymous of this ClusterComputesQuery.  # noqa: E501
        :rtype: bool
        """
        return self._include_anonymous

    @include_anonymous.setter
    def include_anonymous(self, include_anonymous):
        """Sets the include_anonymous of this ClusterComputesQuery.

        Whether to include anonymous Cluster Computes in the search.  # noqa: E501

        :param include_anonymous: The include_anonymous of this ClusterComputesQuery.  # noqa: E501
        :type: bool
        """

        self._include_anonymous = include_anonymous

    @property
    def paging(self):
        """Gets the paging of this ClusterComputesQuery.  # noqa: E501

        Pagination information.  # noqa: E501

        :return: The paging of this ClusterComputesQuery.  # noqa: E501
        :rtype: PageQuery
        """
        return self._paging

    @paging.setter
    def paging(self, paging):
        """Sets the paging of this ClusterComputesQuery.

        Pagination information.  # noqa: E501

        :param paging: The paging of this ClusterComputesQuery.  # noqa: E501
        :type: PageQuery
        """

        self._paging = paging

    @property
    def cloud_id(self):
        """Gets the cloud_id of this ClusterComputesQuery.  # noqa: E501

        Filters Compute Computes by cloud. If this field is absent, no filtering is done.  # noqa: E501

        :return: The cloud_id of this ClusterComputesQuery.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this ClusterComputesQuery.

        Filters Compute Computes by cloud. If this field is absent, no filtering is done.  # noqa: E501

        :param cloud_id: The cloud_id of this ClusterComputesQuery.  # noqa: E501
        :type: str
        """

        self._cloud_id = cloud_id

    @property
    def version(self):
        """Gets the version of this ClusterComputesQuery.  # noqa: E501

        Filters Cluster Computes by version. Versions are positive integers. Setting this field to -1 will return only the latest version of each Cluster Compute. Setting this field to -2 will not filter by version. For example, this can be used to fetch all versions of a Cluster Compute. Deprecated behavior: Setting version to None is equivalent to setting version to '-1'.  # noqa: E501

        :return: The version of this ClusterComputesQuery.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this ClusterComputesQuery.

        Filters Cluster Computes by version. Versions are positive integers. Setting this field to -1 will return only the latest version of each Cluster Compute. Setting this field to -2 will not filter by version. For example, this can be used to fetch all versions of a Cluster Compute. Deprecated behavior: Setting version to None is equivalent to setting version to '-1'.  # noqa: E501

        :param version: The version of this ClusterComputesQuery.  # noqa: E501
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                version is not None and version < -2):  # noqa: E501
            raise ValueError("Invalid value for `version`, must be a value greater than or equal to `-2`")  # noqa: E501

        self._version = version

    @property
    def sort_by_clauses(self):
        """Gets the sort_by_clauses of this ClusterComputesQuery.  # noqa: E501

        The order used to specify results. The list will be used to construct ORDER BY database queries. If not specified, the fallback order by clauses are 1. Creation time (desc) 2. Name (ascending) and 3. ID (ascending)  # noqa: E501

        :return: The sort_by_clauses of this ClusterComputesQuery.  # noqa: E501
        :rtype: list[SortByClauseClusterComputesSortField]
        """
        return self._sort_by_clauses

    @sort_by_clauses.setter
    def sort_by_clauses(self, sort_by_clauses):
        """Sets the sort_by_clauses of this ClusterComputesQuery.

        The order used to specify results. The list will be used to construct ORDER BY database queries. If not specified, the fallback order by clauses are 1. Creation time (desc) 2. Name (ascending) and 3. ID (ascending)  # noqa: E501

        :param sort_by_clauses: The sort_by_clauses of this ClusterComputesQuery.  # noqa: E501
        :type: list[SortByClauseClusterComputesSortField]
        """

        self._sort_by_clauses = sort_by_clauses

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ClusterComputesQuery):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ClusterComputesQuery):
            return True

        return self.to_dict() != other.to_dict()



#!/usr/bin/env python3
"""
Build standalone Anyscale CLI binary.

Creates a self-contained executable with PyInstaller that bundles the Python
interpreter and all dependencies. The binary is architecture-specific but
works across different Python runtime versions.

Usage:
    python3 build_standalone.py
"""

import os
from pathlib import Path
import re
import subprocess
import sys
import tempfile
import time


# Constants
CLI_DIR = Path(__file__).parent
PYINSTALLER_VERSION = "6.18.0"


def get_version():
    """
    Get version from version.py.

    This matches the PyPI package version (e.g., "0.0.0-dev" or "1.2.3").
    """
    version_file = CLI_DIR / "anyscale" / "version.py"
    content = version_file.read_text()
    match = re.search(r'__version__ = ["\']([^"\']+)["\']', content)
    if not match:
        sys.exit("Error: Could not find version in version.py")
    return match.group(1)


def build():
    """Build the standalone binary."""
    start_time = time.time()

    os.chdir(CLI_DIR)

    version = get_version()
    binary_name = f"anyscale-cli-{version}"

    print(f"Building {binary_name}")

    # Install PyInstaller and the anyscale package
    # Package version comes from version.py (no modification needed)
    print("Installing dependencies...")
    try:
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--quiet",
                f"pyinstaller=={PYINSTALLER_VERSION}",
                ".",
            ]
        )
    except subprocess.CalledProcessError as e:
        sys.exit(f"Failed to install dependencies: {e}")

    with tempfile.TemporaryDirectory() as tmpdir:
        # Create symlinks for SDK modules that self-import
        # Generated OpenAPI client code imports itself as top-level modules
        # PyInstaller needs these to resolve imports like "from openapi_client import ..."
        try:
            os.symlink(
                CLI_DIR / "anyscale" / "client" / "openapi_client",
                Path(tmpdir) / "openapi_client",
            )
            os.symlink(
                CLI_DIR / "anyscale" / "sdk" / "anyscale_client",
                Path(tmpdir) / "anyscale_client",
            )
        except OSError as e:
            sys.exit(f"Failed to create symlinks: {e}")

        # Build with PyInstaller
        cmd = [
            sys.executable,
            "-m",
            "PyInstaller",
            "--onefile",  # Create single executable file instead of directory
            "--name",
            binary_name,  # Output binary name
            "--clean",  # Clean cache before building
            "--collect-data",
            "anyscale",  # Include package data files (templates, configs, etc.)
            "--copy-metadata",
            "anyscale",  # Include package metadata for version detection
            "--paths",
            str(CLI_DIR.parent),  # Add search path for finding imports
            "--distpath",
            "dist",  # Output directory for built binary
            "--workpath",
            f"{tmpdir}/build",  # Temporary build directory
            "--specpath",
            tmpdir,  # Location for .spec file (build recipe)
            "anyscale/scripts.py",  # Entry point script
        ]

        # Prepend tmpdir to PYTHONPATH for symlink resolution
        # Keep existing PYTHONPATH so PyInstaller can find installed packages
        env = os.environ.copy()
        existing_pythonpath = env.get("PYTHONPATH", "")
        if existing_pythonpath:
            env["PYTHONPATH"] = f"{tmpdir}:{CLI_DIR.parent}:{existing_pythonpath}"
        else:
            env["PYTHONPATH"] = f"{tmpdir}:{CLI_DIR.parent}"

        try:
            subprocess.check_call(cmd, env=env)
        except subprocess.CalledProcessError as e:
            sys.exit(f"PyInstaller build failed: {e}")

    # Verify binary exists and works
    binary_path = Path("dist") / binary_name
    if not binary_path.exists():
        sys.exit(f"Binary was not created: {binary_path}")

    try:
        result = subprocess.run(
            [str(binary_path), "--version"], capture_output=True, text=True, check=True,
        )
    except subprocess.CalledProcessError as e:
        sys.exit(f"Binary verification failed: {e.stderr}")

    build_time = time.time() - start_time
    print(f"✓ Built: {binary_path} ({binary_path.stat().st_size / 1024 / 1024:.1f} MB)")
    print(f"✓ Version: {result.stdout.strip()}")
    print(f"✓ Build time: {build_time:.1f}s")

    # Return success for shell script
    return 0


if __name__ == "__main__":
    sys.exit(build())
